import logging
import unittest
from pydantic import BaseModel
from hololinked.constants import ResourceTypes
from hololinked.schema_validators.json_schema import JSONSchema
from hololinked.td.data_schema import DataSchema
from hololinked.td.interaction_affordance import (
    PropertyAffordance,
    InteractionAffordance,
    ActionAffordance,
    EventAffordance,
)
from hololinked.core.properties import (
    Property,
    Number,
    String,
    Boolean,
    List,
    Selector,
    ClassSelector,
)
from hololinked.utils import issubklass

try:
    from .things import OceanOpticsSpectrometer, TestThing
    from .utils import TestCase, TestRunner
    from .things.spectrometer import Intensity
except ImportError:
    from things import OceanOpticsSpectrometer, TestThing
    from utils import TestCase, TestRunner
    from things.spectrometer import Intensity


class TestInteractionAffordance(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.thing = OceanOpticsSpectrometer(id="test-thing", log_level=logging.ERROR)
        print(f"Test Interaction Affordance with {cls.__name__}")

    def test_1_associated_objects(self):
        affordance = PropertyAffordance()
        affordance.objekt = OceanOpticsSpectrometer.integration_time
        affordance.owner = self.thing
        # req. 1. internal test for multiple inheritance of pydantic models as there are many classes to track
        self.assertIsInstance(affordance, BaseModel)
        self.assertIsInstance(affordance, DataSchema)
        self.assertIsInstance(affordance, InteractionAffordance)
        self.assertTrue(affordance.what, ResourceTypes.PROPERTY)
        # req. 2. owner must be a Thing
        self.assertEqual(affordance.owner, self.thing)
        # req. 3. when owner is set, thing id & thing class is also set
        self.assertEqual(affordance.thing_id, self.thing.id)
        self.assertEqual(affordance.thing_cls, self.thing.__class__)
        # req. 4. objekt must be a Property, since we use a property affordance here
        self.assertIsInstance(affordance.objekt, Property)
        # req. 5. objekt must be a property of the owner thing
        # --- not enforced yet
        # req. 6. when objekt is set, property name is also set
        self.assertEqual(affordance.name, OceanOpticsSpectrometer.integration_time.name)

        # test the opposite
        affordance = PropertyAffordance()
        # req. 7. accessing any of unset objects should raise an error
        self.assertTrue(affordance.owner is None)
        self.assertTrue(affordance.objekt is None)
        self.assertTrue(affordance.name is None)
        self.assertTrue(affordance.thing_id is None)
        self.assertTrue(affordance.thing_cls is None)

        # req. 8. Only the corresponding object can be set for each affordance type
        # i.e. ActionAffordance accepts only an Action as its Objekt, same for property and same for event
        affordance = ActionAffordance()
        with self.assertRaises(ValueError) as ex:
            affordance.objekt = OceanOpticsSpectrometer.integration_time
        with self.assertRaises(TypeError) as ex:
            affordance.objekt = 5
        self.assertIn(
            "objekt must be instance of Property, Action or Event, given type",
            str(ex.exception),
        )
        affordance.objekt = OceanOpticsSpectrometer.connect
        self.assertTrue(affordance.what, ResourceTypes.ACTION)

        affordance = EventAffordance()
        with self.assertRaises(ValueError) as ex:
            affordance.objekt = OceanOpticsSpectrometer.integration_time
        with self.assertRaises(TypeError) as ex:
            affordance.objekt = 5
        self.assertIn(
            "objekt must be instance of Property, Action or Event, given type",
            str(ex.exception),
        )
        affordance.objekt = OceanOpticsSpectrometer.intensity_measurement_event
        self.assertTrue(affordance.what, ResourceTypes.EVENT)

        affordance = PropertyAffordance()
        with self.assertRaises(ValueError) as ex:
            affordance.objekt = OceanOpticsSpectrometer.connect
        with self.assertRaises(TypeError) as ex:
            affordance.objekt = 5
        self.assertIn(
            "objekt must be instance of Property, Action or Event, given type",
            str(ex.exception),
        )
        affordance.objekt = OceanOpticsSpectrometer.integration_time


class TestDataSchema(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.thing = OceanOpticsSpectrometer(id="test-thing", log_level=logging.ERROR)
        print(f"Test Data Schema with {cls.__name__}")

    """
    OceanOpticsSpectrometer.trigger_mode # selector 
    OceanOpticsSpectrometer.integration_time # number
    OceanOpticsSpectrometer.serial_number # string
    OceanOpticsSpectrometer.nonlinearity_correction # boolean
    OceanOpticsSpectrometer.custom_background_intensity # typed list float, int
    OceanOpticsSpectrometer.wavelengths # list float int
    """

    def test_2_number_schema(self):
        # test implicit generation before actual testing
        schema = OceanOpticsSpectrometer.integration_time.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        self.assertEqual(schema.type, "number")
        # this is because we will use a Property directly so that we can generate dataschema
        # based on different parameters of the property. See below

        integration_time = Number(
            bounds=(1, 1000),
            default=100,
            crop_to_bounds=True,
            step=1,
            doc="integration time in milliseconds",
            metadata=dict(unit="ms"),
        )
        integration_time.__set_name__(OceanOpticsSpectrometer, "integration_time")
        # req. 1. Schema can be created
        schema = integration_time.to_affordance(owner_inst=self.thing)
        # print(schema.json())
        self.assertIsInstance(schema, PropertyAffordance)
        self.assertEqual(schema.type, "number")
        # req. 2. Test number schema specific attributes
        # minimum, maximum, multipleOf
        self.assertEqual(schema.minimum, integration_time.bounds[0])
        self.assertEqual(schema.maximum, integration_time.bounds[1])
        self.assertEqual(schema.multipleOf, integration_time.step)
        self.assertRaises(AttributeError, lambda: schema.exclusiveMinimum)
        self.assertRaises(AttributeError, lambda: schema.exclusiveMaximum)
        # exclusiveMinimum, exclusiveMaximum
        integration_time.inclusive_bounds = (False, False)
        integration_time.step = None
        schema = integration_time.to_affordance(owner_inst=self.thing)
        self.assertEqual(schema.exclusiveMinimum, integration_time.bounds[0])
        self.assertEqual(schema.exclusiveMaximum, integration_time.bounds[1])
        self.assertRaises(AttributeError, lambda: schema.minimum)
        self.assertRaises(AttributeError, lambda: schema.maximum)
        self.assertRaises(AttributeError, lambda: schema.multipleOf)
        # req. 3. oneOf for allow_None to be True
        integration_time.allow_None = True
        schema = integration_time.to_affordance(owner_inst=self.thing)
        self.assertTrue(any(subtype["type"] == "null" for subtype in schema.oneOf))
        self.assertTrue(any(subtype["type"] == "number" for subtype in schema.oneOf))
        self.assertTrue(len(schema.oneOf), 2)
        self.assertTrue(not hasattr(schema, "type") or schema.type is None)
        # when oneOf was used, make sure the entire dataschema is found within the number subtype
        number_schema = next(subtype for subtype in schema.oneOf if subtype["type"] == "number")
        self.assertEqual(number_schema["exclusiveMinimum"], integration_time.bounds[0])
        self.assertEqual(number_schema["exclusiveMaximum"], integration_time.bounds[1])
        self.assertRaises(KeyError, lambda: number_schema["minimum"])
        self.assertRaises(KeyError, lambda: number_schema["maximum"])
        self.assertRaises(KeyError, lambda: number_schema["multipleOf"])
        # print(schema.json())
        # Test some standard data schema values
        self.assertEqual(schema.default, integration_time.default)
        self.assertEqual(schema.unit, integration_time.metadata["unit"])

    def test_3_string_schema(self):
        # test implicit generation before actual testing
        schema = OceanOpticsSpectrometer.status.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)

        status = String(
            regex=r"^[a-zA-Z0-9]{1,10}$",
            default="IDLE",
            doc="status of the spectrometer",
        )
        status.__set_name__(OceanOpticsSpectrometer, "status")
        # req. 1. Schema can be created from the string property
        schema = status.to_affordance(owner_inst=self.thing)
        # print(schema.json())
        self.assertIsInstance(schema, PropertyAffordance)
        self.assertEqual(schema.type, "string")
        # req. 2. Test string schema specific attributes
        self.assertEqual(schema.pattern, status.regex)
        # req. 3. oneOf for allow_None to be True
        status.allow_None = True
        schema = status.to_affordance(owner_inst=self.thing)
        self.assertTrue(any(subtype["type"] == "null" for subtype in schema.oneOf))
        self.assertTrue(any(subtype["type"] == "string" for subtype in schema.oneOf))
        self.assertTrue(len(schema.oneOf), 2)
        self.assertTrue(not hasattr(schema, "type") or schema.type is None)
        # when oneOf was used, make sure the entire dataschema is found within the string subtype
        string_schema = next(subtype for subtype in schema.oneOf if subtype["type"] == "string")
        self.assertEqual(string_schema["pattern"], status.regex)
        # print(schema.json())
        # Test some standard data schema values
        self.assertEqual(schema.default, status.default)

    def test_4_boolean_schema(self):
        # req. 1. Schema can be created from the boolean property and is a boolean schema based property affordance
        schema = OceanOpticsSpectrometer.nonlinearity_correction.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)

        nonlinearity_correction = Boolean(default=True, doc="nonlinearity correction enabled")
        nonlinearity_correction.__set_name__(OceanOpticsSpectrometer, "nonlinearity_correction")
        schema = nonlinearity_correction.to_affordance(owner_inst=self.thing)
        # print(schema.json())
        self.assertIsInstance(schema, PropertyAffordance)
        self.assertEqual(schema.type, "boolean")
        # req. 2. Test boolean schema specific attributes
        # None exists for boolean schema
        # req. 3. oneOf for allow_None to be True
        nonlinearity_correction.allow_None = True
        schema = nonlinearity_correction.to_affordance(owner_inst=self.thing)
        self.assertTrue(any(subtype["type"] == "null" for subtype in schema.oneOf))
        self.assertTrue(any(subtype["type"] == "boolean" for subtype in schema.oneOf))
        self.assertTrue(len(schema.oneOf), 2)
        self.assertTrue(not hasattr(schema, "type") or schema.type is None)
        # print(schema.json())
        # Test some standard data schema values
        self.assertEqual(schema.default, nonlinearity_correction.default)

    def test_5_array_schema(self):
        schema = OceanOpticsSpectrometer.wavelengths.to_affordance(owner_inst=self.thing)
        assert isinstance(schema, PropertyAffordance)

        wavelengths = List(
            default=[],
            item_type=(float, int),
            readonly=True,
            allow_None=False,
            doc="wavelength bins of measurement",
        )
        wavelengths.__set_name__(OceanOpticsSpectrometer, "wavelengths")
        schema = wavelengths.to_affordance(owner_inst=self.thing)
        # req. 1. Schema can be created from the array property and is a array schema based property affordance
        self.assertIsInstance(schema, BaseModel)
        self.assertIsInstance(schema, DataSchema)
        self.assertIsInstance(schema, PropertyAffordance)
        self.assertEqual(schema.type, "array")
        # req. 2. Test array schema specific attributes
        for types in schema.items["oneOf"]:
            self.assertTrue(types["type"] == "number" or types["type"] == "integer")
        # req. 3. Test some standard data schema values
        if OceanOpticsSpectrometer.wavelengths.default is not None:
            self.assertEqual(schema.default, OceanOpticsSpectrometer.wavelengths.default)
        # req. 4. oneOf for allow_None to be True
        OceanOpticsSpectrometer.wavelengths.allow_None = True
        schema = OceanOpticsSpectrometer.wavelengths.to_affordance(owner_inst=self.thing)
        self.assertTrue(any(subtype["type"] == "null" for subtype in schema.oneOf))
        self.assertTrue(any(subtype["type"] == "array" for subtype in schema.oneOf))
        self.assertTrue(len(schema.oneOf), 2)
        self.assertTrue(not hasattr(schema, "type") or schema.type is None)
        # when oneOf was used, make sure the entire dataschema is found within the array subtype
        array_schema = next(subtype for subtype in schema.oneOf if subtype["type"] == "array")
        for types in array_schema["items"]["oneOf"]:  # we know that there are two item types in this array
            self.assertTrue(types["type"] == "number" or types["type"] == "integer")
        # req. 5 check for length constraints
        for bounds in [(5, 1000), (None, 100), (50, None), (51, 101)]:
            wavelengths.bounds = bounds
            wavelengths.allow_None = False
            schema = wavelengths.to_affordance(owner_inst=self.thing)
            if bounds[0] is not None:
                self.assertEqual(schema.minItems, bounds[0])
            else:
                self.assertTrue(not hasattr(schema, "minItems") or schema.minItems is None)
            if bounds[1] is not None:
                self.assertEqual(schema.maxItems, bounds[1])
            else:
                self.assertTrue(not hasattr(schema, "maxItems") or schema.maxItems is None)
            # check if min & max items within allow_None and oneOf
            wavelengths.bounds = bounds
            wavelengths.allow_None = True
            schema = wavelengths.to_affordance(owner_inst=self.thing)
            subtype = next(subtype for subtype in schema.oneOf if subtype["type"] == "array")
            if bounds[0] is not None:
                self.assertEqual(subtype["minItems"], bounds[0])
            else:
                self.assertRaises(KeyError, lambda: subtype["minItems"])
            if bounds[1] is not None:
                self.assertEqual(subtype["maxItems"], bounds[1])
            else:
                self.assertRaises(KeyError, lambda: subtype["maxItems"])

    def test_6_enum_schema(self):
        schema = OceanOpticsSpectrometer.trigger_mode.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)

        trigger_mode = Selector(
            objects=[0, 1, 2, 3, 4],
            default=0,
            observable=True,
            doc="""0 = normal/free running, 1 = Software trigger, 2 = Ext. Trigger Level,
                        3 = Ext. Trigger Synchro/ Shutter mode, 4 = Ext. Trigger Edge""",
        )
        trigger_mode.__set_name__(OceanOpticsSpectrometer, "trigger_mode")
        schema = trigger_mode.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        self.assertEqual(schema.type, "integer")
        self.assertEqual(schema.default, 0)
        # check if enum is equal to objects
        self.assertEqual(schema.enum, trigger_mode.objects)

        # check if allow_None is handled
        trigger_mode.allow_None = True
        trigger_mode.default = 3
        trigger_mode.objects = [0, 1, 2, 3, 4, "0", "1", "2", "3", "4"]
        schema = trigger_mode.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        self.assertTrue(not hasattr(schema, "type") or schema.type is None)
        self.assertEqual(schema.default, 3)
        enum_subschema = next(
            subtype
            for subtype in schema.oneOf
            if (subtype.get("type", None) != "null" or len(subtype.get("oneOf", [])) > 1)
        )
        self.assertIsInstance(enum_subschema, dict)
        self.assertEqual(enum_subschema["enum"], trigger_mode.objects)

    def test_7_class_selector_custom_schema(self):
        last_intensity = ClassSelector(
            default=Intensity([], []),
            allow_None=False,
            class_=Intensity,
            doc="last measurement intensity (in arbitrary units)",
        )
        last_intensity.__set_name__(OceanOpticsSpectrometer, "last_intensity")
        schema = last_intensity.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        # Intensity contains an object schema
        self.assertEqual(schema.type, "object")
        self.assertEqual(schema.properties, Intensity.schema["properties"])

        last_intensity.allow_None = True
        schema = last_intensity.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        self.assertTrue(not hasattr(schema, "type") or schema.type is None)
        subschema = next(subtype for subtype in schema.oneOf if subtype.get("type", None) == "object")
        self.assertIsInstance(subschema, dict)
        self.assertTrue(subschema["type"], "object")
        self.assertEqual(subschema["properties"], Intensity.schema["properties"])

    def test_8_json_schema_properties(self):
        # req. 1. test if all values of a model are found in the property affordance schema
        json_schema_prop = TestThing.json_schema_prop  # type: Property
        json_schema_prop.allow_None = False
        schema = json_schema_prop.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        for key in json_schema_prop.model:
            self.assertEqual(getattr(schema, key, NotImplemented), json_schema_prop.model[key])

        # req. 2. test the schema even if allow None is True
        json_schema_prop.allow_None = True
        schema = json_schema_prop.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        subschema = next(
            subtype
            for subtype in schema.oneOf
            if (subtype.get("type", None) != "null" or len(subtype.get("oneOf", [])) > 1)
        )
        self.assertIsInstance(subschema, dict)
        for key in json_schema_prop.model:
            self.assertEqual(subschema.get(key, NotImplemented), json_schema_prop.model[key])

    def test_9_pydantic_properties(self):
        # req. 1. test if all values of a model are found in the property affordance schema for a BaseModel
        pydantic_prop = TestThing.pydantic_prop  # type: Property
        pydantic_prop.allow_None = False
        schema = pydantic_prop.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        # TODO, this is an inherently harder test case
        if issubklass(pydantic_prop.model, BaseModel):
            self.assertEqual(schema.type, "object")
            for field in pydantic_prop.model.model_fields:
                self.assertIn(field, schema.properties)

        # req. 2 test if all values of a model are found in the property affordance for a BaseModel when allow_None = True
        pydantic_prop.allow_None = True
        schema = pydantic_prop.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        subschema = next(subtype for subtype in schema.oneOf if subtype.get("type", None) == "object")
        self.assertIsInstance(subschema, dict)
        for key in pydantic_prop.model.model_fields:
            self.assertIn(key, subschema.get("properties", {}))

        # req. 3. test if base python types can be used in pydantic property
        pydantic_simple_prop = TestThing.pydantic_simple_prop  # type: Property # its an integer
        pydantic_simple_prop.allow_None = False
        schema = pydantic_simple_prop.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        self.assertEqual(schema.type, "integer")

        pydantic_simple_prop.allow_None = True
        schema = pydantic_simple_prop.to_affordance(owner_inst=self.thing)
        self.assertIsInstance(schema, PropertyAffordance)
        subschema = next(subtype for subtype in schema.oneOf if subtype.get("type", None) == "integer")
        self.assertEqual(subschema["type"], "integer")
        subschema = next(subtype for subtype in schema.oneOf if subtype.get("type", None) == "null")
        self.assertEqual(subschema["type"], "null")


class TestThingDescription(TestCase):
    def test_1_thing_model_generation(self):
        thing = TestThing(id="test-thing-model", log_level=logging.ERROR + 10)
        self.assertIsInstance(thing.get_thing_model(skip_names=["base_property"]).json(), dict)


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestInteractionAffordance))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestDataSchema))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestThingDescription))
    return suite


if __name__ == "__main__":
    unittest.main(testRunner=TestRunner())



import typing
import msgspec
from uuid import uuid4

from ...constants import JSON, byte_types
from ...serializers.serializers import Serializers
from ...serializers.payloads import SerializableData, PreserializedData
from ...param.parameters import Integer

# message types
# both directions
HANDSHAKE = "HANDSHAKE"  # 1 - find out if the server is alive/connect to it
# client to server
OPERATION = "OPERATION"  # 2 - i.e. message type is a request to perform an operation on the interaction affordance
EXIT = "EXIT"  # 3 - exit the server
# server to client
REPLY = "REPLY"  # 4 - response for operation
TIMEOUT = "TIMEOUT"  # 5 - timeout message, operation could not be completed
ERROR = "EXCEPTION"  # 6 - exception occurred while executing operation
INVALID_MESSAGE = "INVALID_MESSAGE"  # 7 - invalid message
SERVER_DISCONNECTED = "EVENT_DISCONNECTED"  # 8 - socket died - zmq's builtin event EVENT_DISCONNECTED
# peer to peer
INTERRUPT = "INTERRUPT"  # 9 - interrupt a socket while polling

# not used now
EVENT = "EVENT"
EVENT_SUBSCRIPTION = "EVENT_SUBSCRIPTION"
SUCCESS = "SUCCESS"


EMPTY_BYTE = b""

"""
Message indices 

| Index | 0       | 1          | 2      | 3      | 4                     |
|-------|---------|------------|--------|--------|-----------------------|
| Desc  | address | empty byte | header | payload| preserialized payload |

"""
# CM = Client Message
INDEX_ADDRESS = 0
INDEX_DELIMITER = 1
INDEX_HEADER = 2
INDEX_BODY = 3
INDEX_PRESERIALIZED_BODY = 4


class ServerExecutionContext(msgspec.Struct):
    """Additional context for the server while executing an operation"""

    invokationTimeout: float
    executionTimeout: float
    oneway: bool


class ThingExecutionContext(msgspec.Struct):
    """Additional context for the thing while executing an operation"""

    fetchExecutionLogs: bool


default_server_execution_context = ServerExecutionContext(invokationTimeout=5, executionTimeout=5, oneway=False)

default_thing_execution_context = ThingExecutionContext(fetchExecutionLogs=False)

SerializableNone = SerializableData(None, content_type="application/json")
PreserializedEmptyByte = PreserializedData(EMPTY_BYTE, content_type="text/plain")


class RequestHeader(msgspec.Struct):
    """Header of a request message"""

    messageType: str
    messageID: str
    senderID: str
    receiverID: str
    serverExecutionContext: ServerExecutionContext = msgspec.field(
        default_factory=lambda: default_server_execution_context
    )
    thingExecutionContext: ThingExecutionContext = msgspec.field(
        default_factory=lambda: default_thing_execution_context
    )
    thingID: typing.Optional[str] = ""
    objekt: typing.Optional[str] = ""
    operation: typing.Optional[str] = ""
    payloadContentType: typing.Optional[str] = "application/json"
    preencodedPayloadContentType: typing.Optional[str] = "text/plain"

    def __getitem__(self, key: str) -> typing.Any:
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f"key {key} not found in {self.__class__.__name__}") from None

    def json(self):
        return {f: getattr(self, f) for f in self.__struct_fields__}


class ResponseHeader(msgspec.Struct):
    """Header of a response message"""

    messageType: str
    messageID: str
    receiverID: str
    senderID: str
    payloadContentType: typing.Optional[str] = "application/json"
    preencodedPayloadContentType: typing.Optional[str] = ""

    def __getitem__(self, key: str) -> typing.Any:
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f"key {key} not found in {self.__class__.__name__}") from None

    def json(self):
        return {f: getattr(self, f) for f in self.__struct_fields__}


class EventHeader(msgspec.Struct):
    """Header of an event message"""

    messageType: str
    messageID: str
    senderID: str
    eventID: str
    payloadContentType: typing.Optional[str] = "application/json"
    preencodedPayloadContentType: typing.Optional[str] = ""

    def __getitem__(self, key: str) -> typing.Any:
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f"key {key} not found in {self.__class__.__name__}") from None

    def json(self):
        return {f: getattr(self, f) for f in self.__struct_fields__}


class RequestMessage:
    """
    A single unit of message from a ZMQ client to server. The message may be parsed and deserialized into header and body.

    Message indices:

    | Index | 0       | 1          | 2      |   3     |          4            |
    |-------|---------|------------|--------|---------|-----------------------|
    | Desc  | address | empty byte | header | payload | preserialized payload |

    For header's JSON schema, visit [here](https://github.com/hololinked-dev/hololinked/blob/main/hololinked/core/zmq/request_message_header_schema.json).
    """

    length = Integer(default=5, readonly=True, class_member=True, doc="length of the message")  # type: int

    def __init__(self, msg: typing.List[bytes]) -> None:
        self._bytes = msg
        self._header = None  # deserialized header
        self._body = None  # type: typing.Optional[typing.Tuple[SerializableData, PreserializedData]]
        self._sender_id = None

    @property
    def byte_array(self) -> typing.List[bytes]:
        """
        message byte array, either after being composed or as received from the socket.

        Message indices:

        | Index | 0       | 1          | 2      |   3     |          4            |
        |-------|---------|------------|--------|---------|-----------------------|
        | Desc  | address | empty byte | header | payload | preserialized payload |
        """
        return self._bytes

    @property
    def header(self) -> RequestHeader:
        """header of the message, namely index 1 of the byte array, deserizalized to a dictionary"""
        if self._header is None:
            self.parse_header()
        return self._header

    @property
    def body(self) -> typing.Tuple[SerializableData, PreserializedData]:
        """body of the message"""
        if self._body is None:
            self.parse_body()
        return self._body

    @property
    def id(self) -> str:
        """ID of the message"""
        return self.header["messageID"]

    @property
    def receiver_id(self) -> str:
        """ID of the sender"""
        return self.header["receiverID"]

    @property
    def sender_id(self) -> str:
        """ID of the receiver"""
        return self.header["senderID"]

    @property
    def thing_id(self) -> str:
        """ID of the thing on which the operation is to be performed"""
        return self.header["thingID"]

    @property
    def type(self) -> str:
        """type of the message"""
        return self.header["messageType"]

    @property
    def server_execution_context(self) -> typing.Dict[str, typing.Any]:
        """server execution context"""
        return self.header["serverExecutionContext"]

    @property
    def thing_execution_context(self) -> typing.Dict[str, typing.Any]:
        """thing execution context"""
        return self.header["thingExecutionContext"]

    @property
    def qualified_operation(self) -> str:
        """qualified objekt - a possibly unique string for the operation"""
        return f"{self.header['thingID']}.{self.header['objekt']}.{self.header['operation']}"

    def parse_header(self) -> None:
        """extract the header and deserialize it"""
        if isinstance(self._bytes[INDEX_HEADER], RequestHeader):
            self._header = self._bytes[INDEX_HEADER]
        elif isinstance(self._bytes[INDEX_HEADER], byte_types):
            self._header = RequestHeader(**Serializers.json.loads(self._bytes[INDEX_HEADER]))
        else:
            raise ValueError(f"header must be of type RequestHeader or bytes, not {type(self._bytes[INDEX_HEADER])}")

    def parse_body(self) -> None:
        """extract the body and deserialize payload"""
        self._body = [
            SerializableData(self._bytes[INDEX_BODY], content_type=self.header["payloadContentType"]),
            PreserializedData(
                self._bytes[INDEX_PRESERIALIZED_BODY], content_type=self.header["preencodedPayloadContentType"]
            ),
        ]

    @classmethod
    def craft_from_arguments(
        cls,
        receiver_id: str,
        sender_id: str,
        thing_id: str,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: typing.Dict[str, typing.Any] = default_server_execution_context,
        thing_execution_context: typing.Dict[str, typing.Any] = default_thing_execution_context,
    ) -> "RequestMessage":
        """
        create a request message from the given arguments

        Parameters
        ----------
        receiver_id: str
            id of the server (ZMQ socket identity)
        sender_id: str
            id of the client (ZMQ socket identity)
        thing_id: str
            id of the thing to which the operation is to be performed
        objekt: str
            objekt of the thing on which the operation is to be performed, i.e. a property, action or event name
        operation: str
            operation to be performed (`invokeaction`, `readproperty`, `writeproperty` etc.)
        payload: SerializableData
            payload for the operation
        preserialized_payload: PreserializedData
            pre-encoded payload for the operation
        server_execution_context: Dict[str, Any]
            server-level execution context while performing the operation
        thing_execution_context: Dict[str, Any]
            thing-level execution context while performing the operation

        Returns
        -------
        message: RequestMessage
            the crafted message
        """
        message = RequestMessage([])
        message._header = RequestHeader(
            messageID=str(uuid4()),
            messageType=OPERATION,
            senderID=sender_id,
            receiverID=receiver_id,
            # i.e. the message type is 'OPERATION', not 'HANDSHAKE', 'REPLY', 'TIMEOUT' etc.
            serverExecutionContext=server_execution_context,
            thingID=thing_id,
            objekt=objekt,
            operation=operation,
            payloadContentType=payload.content_type,
            preencodedPayloadContentType=preserialized_payload.content_type,
            thingExecutionContext=thing_execution_context,
        )
        message._body = [payload, preserialized_payload]
        message._bytes = [
            bytes(receiver_id, encoding="utf-8"),
            bytes(),
            Serializers.json.dumps(message._header.json()),
            payload.serialize(),
            preserialized_payload.value,
        ]
        return message

    @classmethod
    def craft_with_message_type(
        cls, sender_id: str, receiver_id: str, message_type: bytes = HANDSHAKE
    ) -> "RequestMessage":
        """
        create a plain message with a certain type, for example a handshake message.

        Parameters
        ----------
        sender_id: str
            id of the client (ZMQ socket identity)
        receiver_id: str
            id of the server (ZMQ socket identity)
        message_type: bytes
            message type to be sent (i.e. 'HANDSHAKE', 'EXIT' etc.)

        Returns
        -------
        message: RequestMessage
            the crafted message
        """

        message = RequestMessage([])
        message._header = RequestHeader(
            messageID=str(uuid4()),
            messageType=message_type,
            senderID=sender_id,
            receiverID=receiver_id,
            serverExecutionContext=default_server_execution_context,
        )
        payload = SerializableNone
        preserialized_payload = PreserializedEmptyByte
        message._body = [payload, preserialized_payload]
        message._bytes = [
            bytes(receiver_id, encoding="utf-8"),
            bytes(),
            Serializers.json.dumps(message._header.json()),
            payload.serialize(),
            preserialized_payload.value,
        ]
        return message

    def __str__(self) -> str:
        return f"RequestMessage(id={self.id}, type={self.type}, header={self.header})"


class ResponseMessage:
    """
    A single unit of message from a ZMQ server to client.
    The message may be parsed and deserialized into header and body.

    Message indices:

    | Index | 0       |   2    | 3    |     4            |
    |-------|---------|--------|------|------------------|
    | Desc  | address | header | data | pre encoded data |

    For header's JSON schema, visit [here](https://github.com/hololinked-dev/hololinked/blob/main/hololinked/core/zmq/response_message_header_schema.json).
    """

    length = Integer(default=5, readonly=True, class_member=True, doc="length of the message")  # type: int

    def __init__(self, msg: typing.List[bytes]):
        self._bytes = msg
        self._header = None
        self._body = None
        self._sender_id = None

    @property
    def byte_array(self) -> typing.List[bytes]:
        """the message in bytes, either after being composed or as received from the socket.

        Message indices:

        | Index | 0       |   2    | 3    |     4            |
        |-------|---------|--------|------|------------------|
        | Desc  | address | header | data | pre encoded data |
        """
        return self._bytes

    @property
    def id(self) -> str:
        """ID of the message"""
        return self.header["messageID"]

    @property
    def type(self) -> str:
        """type of the message"""
        return self.header["messageType"]

    @property
    def receiver_id(self) -> str:
        """ID of the sender"""
        return self.header["receiverID"]

    @property
    def sender_id(self) -> str:
        """ID of the receiver"""
        return self.header["senderID"]

    @property
    def header(self) -> JSON:
        """header of the message"""
        if self._header is None:
            self.parse_header()
        return self._header

    @property
    def body(self) -> typing.Tuple[SerializableData, PreserializedData]:
        """body of the message"""
        if self._body is None:
            self.parse_body()
        return self._body

    @property
    def payload(self) -> SerializableData:
        """payload of the message"""
        return self.body[0]

    @property
    def preserialized_payload(self) -> PreserializedData:
        """pre-encoded payload of the message"""
        return self.body[1]

    def parse_header(self) -> None:
        """parse the header"""
        if isinstance(self._bytes[INDEX_HEADER], ResponseHeader):
            self._header = self._bytes[INDEX_HEADER]
        elif isinstance(self._bytes[INDEX_HEADER], byte_types):
            self._header = ResponseHeader(**Serializers.json.loads(self._bytes[INDEX_HEADER]))
        else:
            raise ValueError(f"header must be of type ResponseHeader or bytes, not {type(self._bytes[INDEX_HEADER])}")

    def parse_body(self) -> None:
        """parse the body"""
        self._body = [
            SerializableData(self._bytes[INDEX_BODY], content_type=self.header["payloadContentType"]),
            PreserializedData(
                self._bytes[INDEX_PRESERIALIZED_BODY], content_type=self.header["preencodedPayloadContentType"]
            ),
        ]

    @classmethod
    def craft_from_arguments(
        cls,
        receiver_id: str,
        sender_id: str,
        message_type: str,
        message_id: bytes = b"",
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
    ) -> "ResponseMessage":
        """
        Crafts an arbitrary response to the client using the method's arguments.

        Parameters
        ----------
        receiver_id: str
            id of the client (ZMQ socket identity)
        sender_id: str
            id of the server (ZMQ socket identity)
        message_type: str
            type of the message, possible values are 'REPLY', 'HANDSHAKE' and 'TIMEOUT'
        message_id: bytes
            message id of the original client message for which the response is being crafted
        payload: SerializableData
            response payload to send to the client
        preserialized_payload: bytes
            pre-encoded data, generally used for large or custom data that is already serialized

        Returns
        -------
        ResponseMessage
            the crafted response
        """
        message = ResponseMessage([])
        message._header = ResponseHeader(
            messageType=message_type,
            messageID=message_id,
            receiverID=receiver_id,
            senderID=sender_id,
            payloadContentType=payload.content_type,
            preencodedPayloadContentType=preserialized_payload.content_type,
        )
        message._body = [payload, preserialized_payload]
        message._bytes = [
            bytes(receiver_id, encoding="utf-8"),
            bytes(),
            Serializers.json.dumps(message._header.json()),
            payload.serialize(),
            preserialized_payload.value,
        ]
        return message

    @classmethod
    def craft_reply_from_request(
        cls,
        request_message: RequestMessage,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
    ) -> "ResponseMessage":
        """
        Craft a response with certain data extracted from an originating client message,
        like the client's address, message id etc.

        Parameters
        ----------
        request_message: RequestMessage
            The message originated by the client for which the response is being crafted
        payload: SerializableData
            response payload to send to the client
        preserialized_payload: PreserializedData
            pre-encoded data, generally used for large or custom data that is already serialized

        Returns
        -------
        ResponseMessage
            the crafted response
        """
        message = ResponseMessage([])
        message._header = ResponseHeader(
            messageType=REPLY,
            messageID=request_message.id,
            receiverID=request_message.sender_id,
            senderID=request_message.receiver_id,
            payloadContentType=payload.content_type,
            preencodedPayloadContentType=preserialized_payload.content_type,
        )
        message._body = [payload, preserialized_payload]
        message._bytes = [
            bytes(request_message.sender_id, encoding="utf-8"),
            bytes(),
            Serializers.json.dumps(message._header.json()),
            payload.serialize(),
            preserialized_payload.value,
        ]
        return message

    @classmethod
    def craft_with_message_type(
        cls,
        request_message: RequestMessage,
        message_type: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
    ) -> "ResponseMessage":
        """
        create a plain message with a certain type, for example a handshake message.

        Parameters
        ----------
        request_message: RequestMessage
            The message originated by the client for which the response is being crafted
        message_type: str
            message type to be sent
        payload: SerializableData
            response payload to send to the client
        preserialized_payload: PreserializedData
            pre-encoded data, generally used for large or custom data that is already serialized

        Returns
        -------
        ResponseMessage
            the crafted response
        """
        message = ResponseMessage([])
        message._header = ResponseHeader(
            messageType=message_type,
            messageID=request_message.id,
            receiverID=request_message.sender_id,
            senderID=request_message.receiver_id,
            payloadContentType=payload.content_type,
            preencodedPayloadContentType=preserialized_payload.content_type,
        )
        message._body = [payload, preserialized_payload]
        message._bytes = [
            bytes(request_message.sender_id, encoding="utf-8"),
            bytes(),
            Serializers.json.dumps(message._header.json()),
            payload.serialize(),
            preserialized_payload.value,
        ]
        return message

    def __str__(self) -> str:
        return f"ResponseMessage(id={self.id}, type={self.type}, header={self.header})"


class EventMessage(ResponseMessage):
    """
    Event message class for handling events in the system.

    Message indices:

    | Index | 0       | 1      |   2     |          3            |
    |-------|---------|--------|---------|-----------------------|
    | Desc  | address | header | payload | preserialized payload |

    """

    # For header's JSON schema, visit [here](https://hololinked.readthedocs.io/en/latest/protocols/zmq/event-message-header.json).

    @classmethod
    def craft_from_arguments(
        cls,
        event_id: str,
        sender_id: str,
        message_type: str = EVENT,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
    ) -> "EventMessage":
        """
        create a plain message with a certain type, for example a handshake message.

        Parameters
        ----------
        event_id: str
            id of the event (used by ZMQ pub-sub)
        sender_id: str
            id of the sender (ZMQ socket identity)
        message_type: str
            message type to be sent (`EVENT` usually)
        payload: SerializableData
            event payload to send to the client
        preserialized_payload: PreserializedData
            pre-encoded data, generally used for large or custom data that is already serialized

        Returns
        -------
        EventMessage
            the crafted message
        """
        message = EventMessage([])
        message._header = EventHeader(
            messageType=message_type,
            messageID=str(uuid4()),
            eventID=event_id,
            senderID=sender_id,
            payloadContentType=payload.content_type,
            preencodedPayloadContentType=preserialized_payload.content_type,
        )
        message._body = [payload, preserialized_payload]
        message._bytes = [
            bytes(event_id, encoding="utf-8"),
            bytes(),
            Serializers.json.dumps(message._header.json()),
            payload.serialize(),
            preserialized_payload.value,
        ]
        return message

    @property
    def event_id(self) -> str:
        """unique ID of the event by which ZMQ pub-sub works"""
        return self.header["eventID"]

    def parse_header(self) -> None:
        """parse the header"""
        if isinstance(self._bytes[INDEX_HEADER], EventHeader):
            self._header = self._bytes[INDEX_HEADER]
        elif isinstance(self._bytes[INDEX_HEADER], byte_types):
            self._header = EventHeader(**Serializers.json.loads(self._bytes[INDEX_HEADER]))
        else:
            raise ValueError(f"header must be of type ResponseHeader or bytes, not {type(self._bytes[INDEX_HEADER])}")

    def __str__(self) -> str:
        return f"EventMessage(id={self.id}, type={self.type}, header={self.header})"



import copy
import inspect
from types import FunctionType
import typing

from ..param.parameterized import (
    EventResolver as ParamEventResolver,
    EventDispatcher as ParamEventDispatcher,
    Parameter,
    Parameterized,
    ParameterizedMetaclass,
    edit_constant as edit_constant_parameters,
)
from ..utils import getattr_without_descriptor_read
from ..constants import JSON, JSONSerializable
from .actions import Action, BoundAction, action
from .property import Property
from .events import Event, EventPublisher, EventDispatcher


class ThingMeta(ParameterizedMetaclass):
    """
    Metaclass for `Thing`, implements a `__post_init__()` call and instantiation of a registry for properties', actions'
    and events' descriptor objects.
    Accessing properties, actions and events at the class level returns the descriptor object through the `DescriptorRegistry`
    implementation. Accessing properties, actions and events at instance level return their values (for example -
    the value of Property `foo` being `5`).
    `__post_init__()` is run after the user's `__init__()` method, properties that can be
    loaded from a database are loaded and written at this time. Overload `__post_init__()` in subclasses
    to add additional functionality.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/Thing.pdf)
    """

    def __init__(mcs, name, bases, dict_):
        super().__init__(name, bases, dict_)
        mcs._create_actions_registry()
        mcs._create_events_registry()

    def __call__(mcls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)
        instance.__post_init__()
        return instance

    def _create_param_container(cls, cls_members: dict) -> None:
        """
        creates `PropertiesRegistry` instead of `param`'s own `Parameters`
        as the default container for descriptors. All properties have definitions
        copied from `param`.
        """
        cls._param_container = PropertiesRegistry(cls, cls_members)

    def _create_actions_registry(cls) -> None:
        """
        creates `Actions` instead of `param`'s own `Parameters`
        as the default container for descriptors. All actions have definitions
        copied from `param`.
        """
        cls._actions_registry = ActionsRegistry(cls)

    def _create_events_registry(cls) -> None:
        """
        creates `Events` instead of `param`'s own `Parameters`
        as the default container for descriptors. All events have definitions
        copied from `param`.
        """
        cls._events_registry = EventsRegistry(cls)

    @property
    def properties(cls) -> "PropertiesRegistry":
        """
        Container object for Property descriptors. Returns `PropertiesRegistry`
        instance instead of `param`'s own `Parameters` instance.
        """
        return cls._param_container

    @property
    def actions(cls) -> "ActionsRegistry":
        """Container object for Action descriptors"""
        return cls._actions_registry

    @property
    def events(cls) -> "EventsRegistry":
        """Container object for Event descriptors"""
        return cls._events_registry


class DescriptorRegistry:
    """
    A registry for the descriptors of a `Thing` class or `Thing` instance.
    Provides a dictionary interface to access the descriptors under the `descriptors` attribute.
    Each of properties, actions and events subclasss from here to implement a registry of their available objects.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/DescriptorRegistry.pdf)
    """

    def __init__(self, owner_cls: ThingMeta, owner_inst=None) -> None:
        """
        Parameters
        ----------
        owner_cls: ThingMeta
            The class/subclass of the `Thing` that owns the registry.
        owner_inst: Thing
            The instance of the `Thing` that owns the registry, optional
        """
        super().__init__()
        self.owner_cls = owner_cls
        self.owner_inst = owner_inst
        self.clear()

    @property
    def owner(self):
        """
        The owner of the registry - the instance of a `Thing` if a `Thing` has been instantiated
        or the class/subclass of `Thing` when accessed as a class attribute.
        """
        return self.owner_inst if self.owner_inst is not None else self.owner_cls

    @property
    def _qualified_prefix(self) -> str:
        """
        A unique prefix for `descriptors` attribute according to the `Thing`'s subclass and instance id.
        For internal use.
        """
        try:
            return self._qualified__prefix
        except AttributeError:
            prefix = inspect.getfile(self.__class__) + self.__class__.__name__.lower()
            if self.owner_inst is not None:
                prefix += f"_{self.owner_inst.id}"
            self._qualified__prefix = prefix
            return prefix

    @property
    def descriptor_object(self) -> type[Property | Action | Event]:
        """The type of descriptor object that this registry holds, i.e. `Property`, `Action` or `Event`"""
        raise NotImplementedError("Implement descriptor_object in subclass")

    @property
    def descriptors(self) -> typing.Dict[str, type[Property | Action | Event]]:
        """A dictionary with all the descriptors as values and their names as keys."""
        raise NotImplementedError("Implement descriptors in subclass")

    @property
    def names(self) -> typing.KeysView[str]:
        """The names of the descriptors objects as a dictionary key view"""
        return self.descriptors.keys()

    @property
    def values(self) -> typing.Dict[str, typing.Any]:
        """
        The values contained within the descriptors after reading when accessed at instance level, otherwise,
        the descriptor objects as dictionary when accessed at class level.
        """
        raise NotImplementedError("Implement values in subclass")

    def clear(self) -> None:
        """
        Deletes the descriptors dictionary (value of the `descriptors` property) so that it can be recreated.
        Does not delete the descriptors themselves. Call this method once if new descriptors are added to the
        class/instance dynamically in runtime.
        """
        for name in ["", "_values"]:
            try:
                delattr(
                    self,
                    f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}{name}",
                )
            except AttributeError:
                pass

    def __getitem__(self, key: str) -> Property | Action | Event:
        """Returns the descriptor object for the given key."""
        raise NotImplementedError("Implement __getitem__ in subclass")

    def __contains__(self, obj: Property | Action | Event) -> bool:
        """Returns `True` if the descriptor object is in the descriptors dictionary."""
        raise NotImplementedError("contains not implemented yet")

    def __dir__(self) -> typing.List[str]:
        """Adds descriptor object to the dir"""
        return super().__dir__() + self.descriptors.keys()  # type: ignore

    def __iter__(self):
        """Iterates over the descriptors of this object."""
        yield from self.descriptors

    def __len__(self) -> int:
        """The number of descriptors in this object."""
        return len(self.descriptors)

    def __hash__(self) -> int:
        return hash(self._qualified__prefix)

    def __str__(self) -> int:
        if self.owner_inst:
            return f"<DescriptorRegistry({self.owner_cls.__name__}({self.owner_inst.id}))>"
        return f"<DescriptorRegistry({self.owner_cls.__name__})>"

    def get_descriptors(self, recreate: bool = False) -> typing.Dict[str, Property | Action | Event]:
        """
        a dictionary with all the descriptors as values and their names as keys.

        Parameters
        ----------
        recreate: bool
            if `True`, the descriptors dictionary is recreated and returned, otherwise, the cached dictionary is returned.
        """
        if recreate:
            self.clear()
        try:
            return getattr(self, f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}")
        except AttributeError:
            descriptors = dict()
            for name, objekt in inspect._getmembers(
                self.owner_cls,
                lambda f: isinstance(f, self.descriptor_object),
                getattr_without_descriptor_read,
            ):
                descriptors[name] = objekt
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}",
                descriptors,
            )
            # We cache the parameters because this method is called often,
            # and parameters are rarely added (and cannot be deleted)
            return descriptors

    def get_values(self) -> typing.Dict[str, typing.Any]:
        """
        the values contained within the descriptors after reading when accessed at instance level, otherwise,
        the descriptor objects as dictionary when accessed at class level.
        For example, if a `Thing` instance's property contains a value of 5, this method will return
        { property_name : 5 } when accessed at instance level, and { property_name : property_object } when accessed
        at class level.
        This method is also the getter of the `values` property.
        """
        if self.owner_inst is None:
            return self.descriptors
        try:
            return getattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_values",
            )
        except AttributeError:
            values = dict()
            for name, value in self.descriptors.items():
                values[name] = value.__get__(self.owner_inst, self.owner_cls)
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_values",
                values,
            )
            return values


def supports_only_instance_access(
    error_msg: str = "This method is only supported at instance level",
) -> FunctionType:
    """
    decorator to raise an error if a method is called at class level instead of instance level
    within the registry functionality.
    """

    def inner(func: FunctionType) -> FunctionType:
        def wrapper(self: DescriptorRegistry, *args, **kwargs):
            if self.owner_inst is None:
                error_msg = inner._error_msg
                raise AttributeError(error_msg)
            return func(self, *args, **kwargs)

        return wrapper

    inner._error_msg = error_msg
    return inner


class PropertiesRegistry(DescriptorRegistry):
    """
    A `DescriptorRegistry` for properties of a `Thing` class or `Thing` instance.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/DescriptorRegistry.pdf)
    """

    def __init__(self, owner_cls: ThingMeta, owner_class_members: dict, owner_inst=None):
        super().__init__(owner_cls, owner_inst)
        if self.owner_inst is None and owner_class_members is not None:
            # instantiated by class
            self.event_resolver = ParamEventResolver(owner_cls=owner_cls)
            self.event_dispatcher = ParamEventDispatcher(owner_cls, self.event_resolver)
            self.event_resolver.create_unresolved_watcher_info(owner_class_members)
        else:
            # instantiated by instance
            self._instance_params = {}
            self.event_resolver = self.owner_cls.properties.event_resolver
            self.event_dispatcher = ParamEventDispatcher(owner_inst, self.event_resolver)
            self.event_dispatcher.prepare_instance_dependencies()

    @property
    def descriptor_object(self) -> type[Parameter]:
        return Parameter

    @property
    def descriptors(self) -> typing.Dict[str, Parameter]:
        if self.owner_inst is None:
            return super().get_descriptors()
        return dict(super().get_descriptors(), **self._instance_params)

    values = property(DescriptorRegistry.get_values, doc=DescriptorRegistry.get_values.__doc__)  # type: typing.Dict[str, Parameter | Property | typing.Any]

    def __getitem__(self, key: str) -> Property | Parameter:
        return self.descriptors[key]

    def __contains__(self, value: str | Property | Parameter) -> bool:
        return value in self.descriptors.values() or value in self.descriptors

    @property
    def defaults(self) -> typing.Dict[str, typing.Any]:
        """default values of all properties as a dictionary with property names as keys"""
        defaults = {}
        for key, val in self.descriptors.items():
            defaults[key] = val.default
        return defaults

    @property
    def remote_objects(self) -> typing.Dict[str, Property]:
        """
        dictionary of properties that are remotely accessible (`remote=True`),
        which is also a default setting for all properties
        """
        try:
            return getattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_remote",
            )
        except AttributeError:
            props = self.descriptors
            remote_props = {}
            for name, desc in props.items():
                if not isinstance(desc, Property):
                    continue
                if desc.is_remote:
                    remote_props[name] = desc
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_remote",
                remote_props,
            )
            return remote_props

    @property
    def db_objects(self) -> typing.Dict[str, Property]:
        """
        dictionary of properties that are stored or loaded from the database
        (`db_init`, `db_persist` or `db_commit` set to `True`)
        """
        try:
            return getattr(self, f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_db")
        except AttributeError:
            propdict = self.descriptors
            db_props = {}
            for name, desc in propdict.items():
                if not isinstance(desc, Property):
                    continue
                if desc.db_init or desc.db_persist or desc.db_commit:
                    db_props[name] = desc
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_db",
                db_props,
            )
            return db_props

    @property
    def db_init_objects(self) -> typing.Dict[str, Property]:
        """dictionary of properties that are initialized from the database (`db_init` or `db_persist` set to `True`)"""
        try:
            return getattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_db_init",
            )
        except AttributeError:
            propdict = self.db_objects
            db_init_props = {}
            for name, desc in propdict.items():
                if desc.db_init or desc.db_persist:
                    db_init_props[name] = desc
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_db_init",
                db_init_props,
            )
            return db_init_props

    @property
    def db_commit_objects(self) -> typing.Dict[str, Property]:
        """dictionary of properties that are committed to the database (`db_commit` or `db_persist` set to `True`)"""
        try:
            return getattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_db_commit",
            )
        except AttributeError:
            propdict = self.db_objects
            db_commit_props = {}
            for name, desc in propdict.items():
                if desc.db_commit or desc.db_persist:
                    db_commit_props[name] = desc
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_db_commit",
                db_commit_props,
            )
            return db_commit_props

    @property
    def db_persisting_objects(self) -> typing.Dict[str, Property]:
        """dictionary of properties that are persisted through the database (`db_persist` set to `True`)"""
        try:
            return getattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_db_persisting",
            )
        except AttributeError:
            propdict = self.db_objects
            db_persisting_props = {}
            for name, desc in propdict.items():
                if desc.db_persist:
                    db_persisting_props[name] = desc
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_db_persisting",
                db_persisting_props,
            )
            return db_persisting_props

    def get(self, **kwargs) -> typing.Dict[str, typing.Any]:
        """
        read properties from the object, implements WoT operations `readAllProperties` and `readMultipleProperties`

        Parameters
        ----------
        **kwargs: typing.Dict[str, typing.Any]

            - names: `List[str]`
                list of property names to be fetched

            - or, key-value pairs of property
                name of the property to be fetched, along with a 'rename' for the property in the response.
                For example { 'foo_prop' : 'fooProp' } will return the property 'foo_prop' as 'fooProp' in the response.

        Returns
        -------
        typing.Dict[str, typing.Any]
            dictionary of property names and their values

        Raises
        ------
        TypeError
            if property name is not a string or requested new name is not a string
        AttributeError
            if property does not exist or is not remote accessible
        """
        data = {}
        if len(kwargs) == 0:
            # read all properties
            for name, prop in self.remote_objects.items():
                if self.owner_inst is None and not prop.class_member:
                    continue
                data[name] = prop.__get__(self.owner_inst, self.owner_cls)
            return data
        elif "names" in kwargs:
            names = kwargs.get("names")
            if not isinstance(names, (list, tuple, str)):
                raise TypeError(
                    "Specify properties to be fetched as a list, tuple or comma separated names. "
                    + f"Given type {type(names)}"
                )
            if isinstance(names, str):
                names = names.split(",")
            kwargs = {name: name for name in names}
        for requested_prop, rename in kwargs.items():
            if not isinstance(requested_prop, str):
                raise TypeError(f"property name must be a string. Given type {type(requested_prop)}")
            if not isinstance(rename, str):
                raise TypeError(f"requested new name must be a string. Given type {type(rename)}")
            if requested_prop not in self.descriptors:
                raise AttributeError(f"property {requested_prop} does not exist")
            if requested_prop not in self.remote_objects:
                raise AttributeError(f"property {requested_prop} is not remote accessible")
            prop = self.descriptors[requested_prop]
            if self.owner_inst is None and not prop.class_member:
                continue
            data[rename] = prop.__get__(self.owner_inst, self.owner_cls)
        return data

    def set(self, **values: typing.Dict[str, typing.Any]) -> None:
        """
        set properties whose name is specified by keys of a dictionary; implements WoT operations `writeMultipleProperties`
        or `writeAllProperties`.

        Parameters
        ----------
        values: typing.Dict[str, typing.Any]
            dictionary of property names and its new values

        Raises
        ------
        AttributeError
            if property does not exist or is not remote accessible
        """
        errors = ""
        for name, value in values.items():
            try:
                if name not in self.descriptors:
                    raise AttributeError(f"property {name} does not exist")
                if name not in self.remote_objects:
                    raise AttributeError(f"property {name} is not remote accessible")
                prop = self.descriptors[name]
                if self.owner_inst is None and not prop.class_member:
                    raise AttributeError(f"property {name} is not a class member and cannot be set at class level")
                setattr(self.owner, name, value)
            except Exception as ex:
                errors += f"{name}: {str(ex)}\n"
        if errors:
            ex = RuntimeError(
                "Some properties could not be set due to errors. "
                + "Check exception notes or server logs for more information."
            )
            ex.__notes__ = errors
            raise ex from None

    def add(self, name: str, config: JSON) -> None:
        """
        add a property to the `Thing` object

        Parameters
        ----------
        name: str
            name of the property
        config: JSON
            configuration of the property, i.e. keyword arguments to the `__init__` method of the property class
        """
        prop = self.get_type_from_name(**config)
        setattr(self.owner_cls, name, prop)
        prop.__set_name__(self.owner_cls, name)
        if prop.deepcopy_default:
            self._deep_copy_param_descriptor(prop)
            self._deep_copy_param_default(prop)
        self.clear()

    def clear(self):
        super().clear()
        self._instance_params = {}
        for attr in ["_db", "_db_init", "_db_persisting", "_remote"]:
            try:
                delattr(
                    self,
                    f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}{attr}",
                )
            except AttributeError:
                pass

    @supports_only_instance_access("database operations are only supported at instance level")
    def get_from_DB(self) -> typing.Dict[str, typing.Any]:
        """
        get all properties (i.e. their values) currently stored in the database

        Returns
        -------
        Dict[str, typing.Any]
            dictionary of property names and their values
        """
        if not hasattr(self.owner_inst, "db_engine"):
            raise AttributeError("database engine not set, this object is not connected to a database")
        props = self.owner_inst.db_engine.get_all_properties()  # type: typing.Dict
        final_list = {}
        for name, prop in props.items():
            try:
                # serializer = Serializers.for_object(self.owner_inst.id, self.owner_cls.__name__, name)
                # if name in self.db_commit_objects:
                #     continue
                final_list[name] = prop
            except Exception as ex:
                self.owner_inst.logger.error(
                    f"could not deserialize property {name} due to error - {str(ex)}, skipping this property"
                )
        return final_list

    @supports_only_instance_access("database operations are only supported at instance level")
    def load_from_DB(self):
        """
        Load and apply property values from database which have `db_init` or `db_persist` set to `True`
        """
        if not hasattr(self.owner_inst, "db_engine"):
            return
            # raise AttributeError("database engine not set, this object is not connected to a database")
        missing_properties = self.owner_inst.db_engine.create_missing_properties(
            self.db_init_objects, get_missing_property_names=True
        )
        # 4. read db_init and db_persist objects
        with edit_constant_parameters(self.owner_inst):
            for db_prop, value in self.get_from_DB().items():
                try:
                    prop_desc = self.descriptors[db_prop]  # type: Property
                    if (prop_desc.db_init or prop_desc.db_persist) and db_prop not in missing_properties:
                        setattr(self.owner_inst, db_prop, value)  # type: ignore
                except Exception as ex:
                    self.owner_inst.logger.error(f"could not set attribute {db_prop} due to error {str(ex)}")

    @classmethod
    def get_type_from_name(cls, name: str) -> typing.Type[Property]:
        return Property

    @supports_only_instance_access("additional property setup is required only for instances")
    def _setup_parameters(self, **parameters):
        """
        Initialize default and keyword parameter values.

        First, ensures that all Parameters with 'deepcopy_default=True'
        (typically used for mutable Parameters) are copied directly
        into each object, to ensure that there is an independent copy
        (to avoid surprising aliasing errors).  Then sets each of the
        keyword arguments, warning when any of them are not defined as
        parameters.

        Constant Parameters can be set during calls to this method.
        """
        ## Deepcopy all 'deepcopy_default=True' parameters
        # (building a set of names first to avoid redundantly
        # instantiating a later-overridden parent class's parameter)
        param_default_values_to_deepcopy = {}
        param_descriptors_to_deepcopy = {}
        for k, v in self.owner_cls.properties.descriptors.items():
            if v.deepcopy_default and k != "name":
                # (avoid replacing name with the default of None)
                param_default_values_to_deepcopy[k] = v
            if v.per_instance_descriptor and k != "name":
                param_descriptors_to_deepcopy[k] = v

        for p in param_default_values_to_deepcopy.values():
            self._deep_copy_param_default(p)
        for p in param_descriptors_to_deepcopy.values():
            self._deep_copy_param_descriptor(p)

        ## keyword arg setting
        if len(parameters) > 0:
            descs = self.descriptors
            for name, val in parameters.items():
                desc = descs.get(name, None)  # pylint: disable-msg=E1101
                if desc:
                    setattr(self.owner_inst, name, val)
                # Its erroneous to set a non-descriptor (& non-param-descriptor) with a value from init.
                # we dont know what that value even means, so we silently ignore

    @supports_only_instance_access("additional property setup is required only for instances")
    def _deep_copy_param_default(self, param_obj: "Parameter") -> None:
        # deepcopy param_obj.default into self.__dict__ (or dict_ if supplied)
        # under the parameter's _internal_name (or key if supplied)
        _old = self.owner_inst.__dict__.get(param_obj._internal_name, NotImplemented)
        _old = _old if _old is not NotImplemented else param_obj.default
        new_object = copy.deepcopy(_old)
        # remember : simply setting in the dict does not activate post setter and remaining logic which is sometimes important
        self.owner_inst.__dict__[param_obj._internal_name] = new_object

    @supports_only_instance_access("additional property setup is required only for instances")
    def _deep_copy_param_descriptor(self, param_obj: Parameter):
        param_obj_copy = copy.deepcopy(param_obj)
        self._instance_params[param_obj.name] = param_obj_copy


class ActionsRegistry(DescriptorRegistry):
    """
    A `DescriptorRegistry` for actions of a `Thing` class or `Thing` instance.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/DescriptorRegistry.pdf)
    """

    @property
    def descriptor_object(self) -> type[Action]:
        return Action

    descriptors = property(DescriptorRegistry.get_descriptors)  # type: typing.Dict[str, Action]

    values = property(DescriptorRegistry.get_values, doc=DescriptorRegistry.get_values.__doc__)  # type: typing.Dict[str, Action]

    def __getitem__(self, key: str) -> Action | BoundAction:
        if self.owner_inst is not None:
            return self.descriptors[key].__get__(self.owner_inst, self.owner_cls)
        return self.descriptors[key]

    def __contains__(self, action: str | Action | BoundAction) -> bool:
        return action in self.descriptors.values() or action in self.descriptors


class EventsRegistry(DescriptorRegistry):
    """
    A `DescriptorRegistry` for events of a `Thing` class or `Thing` instance.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/DescriptorRegistry.pdf)
    """

    @property
    def descriptor_object(self):
        return Event

    descriptors = property(DescriptorRegistry.get_descriptors)  # type: typing.Dict[str, Event]

    values = property(DescriptorRegistry.get_values, doc=DescriptorRegistry.get_values.__doc__)  # type: typing.Dict[str, EventDispatcher]

    def __getitem__(self, key: str) -> Event | EventDispatcher:
        if self.owner_inst is not None:
            return self.descriptors[key].__get__(self.owner_inst, self.owner_cls)
        return self.descriptors[key]

    def __contains__(self, event: Event) -> bool:
        return event in self.descriptors.values() or event in self.descriptors

    def clear(self):
        super().clear()
        for attr in ["_change_events", "_observables"]:
            try:
                delattr(
                    self,
                    f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}{attr}",
                )
            except AttributeError:
                pass

    @property
    def plain(self) -> typing.Dict[str, Event]:
        """dictionary of events that are not change events (i.e., not observable)"""
        try:
            return getattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_non_change_events",
            )
        except AttributeError:
            non_change_events = dict()
            for name, evt in self.descriptors.items():
                if not evt._observable:
                    non_change_events[name] = evt
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_non_change_events",
                non_change_events,
            )
            return non_change_events

    @property
    def change_events(self) -> typing.Dict[str, Event]:
        """dictionary of change events belonging to observable properties"""
        try:
            return getattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_change_events",
            )
        except AttributeError:
            change_events = dict()
            for name, evt in self.descriptors.items():
                if not evt._observable:
                    continue
                change_events[name] = evt
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_change_events",
                change_events,
            )
            return change_events

    @property
    def observables(self) -> typing.Dict[str, Property]:
        """dictionary of all properties that are observable, i.e. that which push change events"""
        try:
            return getattr(
                self,
                f"_{self._qualified__prefix}_{self.__class__.__name__.lower()}_observables",
            )
        except AttributeError:
            props = dict()
            for name, prop in self.owner_cls.properties.descriptors.items():
                if not isinstance(prop, Property) or not prop.observable:
                    continue
                props[name] = prop
            setattr(
                self,
                f"_{self._qualified_prefix}_{self.__class__.__name__.lower()}_observables",
                props,
            )
            return props


class Propertized(Parameterized):
    """
    Base class providing additional functionality related to properties,
    like setting up a registry, allowing values to be set at `__init__()` etc.
    It is not meant to be subclassed directly by the end-user.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/Thing.pdf)
    """

    # There is a word called Property+ize in english dictionary
    # https://en.wiktionary.org/wiki/propertization

    id: str

    # creating name without underscore causes clash with the metaclass method
    # with same name
    def create_param_container(self, **params):
        self._properties_registry = PropertiesRegistry(self.__class__, None, self)
        self._properties_registry._setup_parameters(**params)
        self._param_container = self._properties_registry  # backwards compatibility with param

    @property
    def properties(self) -> PropertiesRegistry:
        """container for the property descriptors of the object."""
        return self._properties_registry

    # we need to specification define it as an action to for the possibility of getting an
    # Affordance object associated with it i.e _get_properties.to_affordance() function needs to work.
    # TODO - fix this anomaly
    @action()
    def _get_properties(self, **kwargs) -> typing.Dict[str, typing.Any]:
        """ """
        return self.properties.get(**kwargs)

    @action()
    def _set_properties(self, **values: typing.Dict[str, typing.Any]) -> None:
        """
        set properties whose name is specified by keys of a dictionary

        Parameters
        ----------
        values: Dict[str, Any]
            dictionary of property names and its values
        """
        return self.properties.set(**values)  # returns None

    @action()
    def _get_properties_in_db(self) -> typing.Dict[str, JSONSerializable]:
        """
        get all properties in the database

        Returns
        -------
        Dict[str, JSONSerializable]
            dictionary of property names and their values
        """
        return self.properties.get_from_DB()

    @action()
    def _add_property(self, name: str, prop: JSON) -> None:
        """
        add a property to the object

        Parameters
        ----------
        name: str
            name of the property
        prop: Property
            property object
        """
        raise NotImplementedError("this method will be implemented properly in a future release")
        prop = Property(**prop)
        self.properties.add(name, prop)
        self._prepare_resources()
        # instruct the clients to fetch the new resources


class RemoteInvokable:
    """
    Base class providing additional functionality related to actions,
    it is not meant to be subclassed directly by the end-user.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/Thing.pdf)
    """

    id: str

    def __init__(self):
        super().__init__()
        self.create_actions_registry()

    # creating name without underscore causes clash with the metaclass method
    # with same name
    def create_actions_registry(self) -> None:
        """creates a registry for available `Actions` based on `ActionsRegistry`"""
        self._actions_registry = ActionsRegistry(self.__class__, self)

    @property
    def actions(self) -> ActionsRegistry:
        """container for the action descriptors of the object."""
        return self._actions_registry


class EventSource:
    """
    Base class to add event functionality to an object,
    it is not meant to be subclassed directly by the end-user.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/Thing.pdf)
    """

    id: str

    def __init__(self) -> None:
        self.create_events_registry()

    # creating name without underscore causes clash with the metaclass method
    # with same name
    def create_events_registry(self) -> None:
        """creates a registry for available `Events` based on `EventsRegistry`"""
        self._events_registry = EventsRegistry(self.__class__, self)

    @property
    def events(self) -> EventsRegistry:
        """container for the event descriptors of the object."""
        return self._events_registry

    @property
    def event_publisher(self) -> "EventPublisher":
        """
        event publishing object `EventPublisher` that owns the zmq.PUB socket, valid only after
        creating an RPC server or calling a `run()` method on the `Thing` instance.
        """
        try:
            return self.rpc_server.event_publisher if self.rpc_server else None
        except AttributeError:
            return None



import logging
import inspect
import threading
import ssl
import typing
import warnings

from ..constants import ZMQ_TRANSPORTS
from ..utils import *  # noqa: F403
from ..exceptions import *  # noqa: F403
from ..serializers import Serializers, BaseSerializer, JSONSerializer
from ..server.server import BaseProtocolServer  # noqa: F401
from .properties import String, ClassSelector
from .property import Property
from .actions import BoundAction, action
from .events import EventDispatcher
from .meta import ThingMeta, Propertized, RemoteInvokable, EventSource


class Thing(Propertized, RemoteInvokable, EventSource, metaclass=ThingMeta):
    """
    Subclass from here to expose hardware or python objects on the network. Remotely accessible members of a `Thing` are
    segragated into properties, actions & events. Utilize properties for data that can be read and written,
    actions to instruct the object to perform tasks and events to get notified of any relevant information. State Machines
    can be used to contrain operations on properties and actions.

    [UML Diagram](https://docs.hololinked.dev/UML/PDF/Thing.pdf)
    """

    # local properties
    id = String(
        default=None,
        regex=r"[A-Za-z]+[A-Za-z_0-9\-\/]*",
        constant=True,
        remote=False,
        doc="""String identifier of the instance. For an interconnected system of hardware, 
            IDs are recommended to be unique. This value is used for many operations,
            for example - creating zmq socket address, tables in databases, and to identify the instance 
            in the HTTP Server - (http(s)://{domain and sub domain}/{id}).""",
    )  # type: str

    logger = ClassSelector(
        class_=logging.Logger,
        default=None,
        allow_None=True,
        remote=False,
        doc="""logging.Logger instance to track log messages. Default logger with a IO-stream handler 
            and network accessible handler is created if none supplied.""",
    )  # type: logging.Logger

    state_machine = None  # type: typing.Optional["StateMachine"]

    # remote properties
    state = String(
        default=None,
        allow_None=True,
        readonly=True,
        observable=True,
        fget=lambda self: self.state_machine.current_state if self.state_machine else None,
        doc="""current state machine's state if state machine present, `None` indicates absence of state machine.
                State machine returned state is always a string even if specified as an Enum in the state machine.""",
    )  # type: typing.Optional[str]

    # object_info = Property(doc="contains information about this object like the class name, script location etc.") # type: ThingInformation

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        # defines some internal fixed attributes. attributes created by us that require no validation but
        # cannot be modified are called _internal_fixed_attributes
        obj._internal_fixed_attributes = ["_internal_fixed_attributes", "_owners"]
        return obj

    def __init__(
        self,
        *,
        id: str,
        logger: typing.Optional[logging.Logger] = None,
        serializer: typing.Optional[BaseSerializer | JSONSerializer] = None,
        **kwargs: typing.Dict[str, typing.Any],
    ) -> None:
        """
        Parameters
        ----------
        id: str
            String identifier of the instance. For an interconnected system of hardware,
            IDs are recommended to be unique. This value is used for many operations, for example -
            creating zmq socket address, tables in databases, and to identify the instance in a
            HTTP Server - (http(s)://{domain and sub domain}/{id}).
        logger: logging.Logger, optional
            `logging.Logger` instance to track log messages. Default logger with a IO-stream handler
            and network accessible handler is created if None supplied.
        serializer: BaseSerializer | JSONSerializer, optional
            Default serializer to be used for serializing and deserializing data.
            If not supplied, a `msgspec` based JSON Serializer is used.
        **kwargs: typing.Dict[str, Any]
            - `remote_accessible_logger`: `bool`, Default `False`.
                if `True`, the log records can be streamed by a remote client. `remote_accessible_logger` can also be set as a
                class attribute.
            - `use_default_db`: `bool`, Default `False`.
                if `True`, default SQLite database is created where properties can be stored and loaded. There is no need to supply
                any database credentials. `use_default_db` value can also be set as a class attribute.
            - `db_config_file`: `str`, optional.
                if not using a default database, supply a JSON configuration file to create a database connection. Check documentaion
                of `hololinked.core.database`.
            - `use_json_file`: `bool`, Default `False`
                if `True`, a JSON file will be used as the property storage instead of a database. This value can also be
                set as a class attribute.
            - `json_filename`: `str`, optional
                If using JSON storage, this filename is used to persist property values. If not provided, a default filename
                is generated based on the instance name.
        """
        Propertized.__init__(self, id=id, logger=logger, **kwargs)
        RemoteInvokable.__init__(self)
        EventSource.__init__(self)
        if self.id.startswith("/"):
            self.id = self.id[1:]
            self.logger.info("removed leading '/' from id")
        if serializer is not None:
            Serializers.register_for_thing_instance(self.id, serializer)

        from .logger import prepare_object_logger
        from .state_machine import prepare_object_FSM
        from ..storage import prepare_object_storage

        prepare_object_logger(
            instance=self,
            log_level=kwargs.get("log_level", None),
            log_file=kwargs.get("log_file", None),
            remote_access=kwargs.get(
                "remote_accessible_logger",
                self.__class__.remote_accessible_logger
                if hasattr(self.__class__, "remote_accessible_logger")
                else False,
            ),
        )
        prepare_object_FSM(self)
        prepare_object_storage(self, **kwargs)  # use_default_db, db_config_file, use_json_file, json_filename

        self._qualified_id = self.id  # filler for now - TODO
        # thing._qualified_id = f'{self._qualified_id}/{thing.id}'

    def __post_init__(self):
        from .zmq.rpc_server import RPCServer  # noqa: F401
        from ..server.zmq import ZMQServer  # noqa: F401
        from .logger import RemoteAccessHandler
        from ..storage.database import ThingDB

        # Type definitions
        self.rpc_server = None  # type: typing.Optional[RPCServer | ZMQServer]
        self.db_engine: typing.Optional[ThingDB]
        self._owners = None if not hasattr(self, "_owners") else self._owners  # type: typing.Optional[typing.List[Thing]]
        self._remote_access_loghandler: typing.Optional[RemoteAccessHandler]
        self._internal_fixed_attributes: typing.List[str]
        self._qualified_id: str
        self._state_machine_state: str
        # database operations
        self.properties.load_from_DB()
        # object is ready
        self.logger.info(f"initialialised Thing class {self.__class__.__name__} with id {self.id}")

    def __setattr__(self, __name: str, __value: typing.Any) -> None:
        if __name == "_internal_fixed_attributes" or __name in self._internal_fixed_attributes:
            # order of 'or' operation for above 'if' matters
            if not hasattr(self, __name) or getattr(self, __name, None) is None:
                # allow setting of fixed attributes once
                super().__setattr__(__name, __value)
            else:
                raise AttributeError(
                    f"Attempted to set {__name} more than once. "
                    + "Cannot assign a value to this variable after creation."
                )
        else:
            super().__setattr__(__name, __value)

    @property
    def sub_things(self) -> typing.Dict[str, "Thing"]:
        """other `Thing`s' that are composed within this `Thing`."""
        things = dict()
        for name, subthing in inspect._getmembers(
            self,
            lambda obj: isinstance(obj, Thing),
            getattr_without_descriptor_read,  # noqa: F405
        ):
            if not hasattr(subthing, "_owners") or subthing._owners is None:
                subthing._owners = []
            if self not in subthing._owners:
                subthing._owners.append(self)
            things[name] = subthing
        return things

    @action()
    def get_thing_model(self, ignore_errors: bool = False, skip_names: list[str] = []):
        """
        generate the [Thing Model](https://www.w3.org/TR/wot-thing-description11/#introduction-tm) of the object.
        The model is a JSON that describes the object's properties, actions, events and their metadata, without the
        protocol information. The model can be used by a client to understand the object's capabilities.

        Parameters
        ----------
        ignore_errors: bool, optional, Default `False`
            if `True`, offending interaction affordances will be removed from the JSON
            (i.e. those who have wrong metadata or non-JSON metadata).
            This is useful to build partial but always working ThingModel.
        skip_names: list[str], optional
            List of affordances names (of any type) to skip in the generated model.

        Returns
        -------
        hololinked.td.ThingModel
            represented as an object in python, gets automatically serialized to JSON when pushed out of the socket.
        """
        # allow_loose_schema: bool, optional, Default False
        #     Experimental properties, actions or events for which schema was not given will be supplied with a suitable
        #     value for node-wot to ignore validation or claim the accessed value for complaint with the schema.
        #     In other words, schema validation will always pass.
        from ..td.tm import ThingModel

        return ThingModel(instance=self, ignore_errors=ignore_errors, skip_names=skip_names).generate()

    thing_model = property(get_thing_model, doc=get_thing_model.__doc__)

    @forkable  # noqa: F405
    def run_with_zmq_server(
        self,
        access_points: list[ZMQ_TRANSPORTS] | ZMQ_TRANSPORTS | str | list[str] = ZMQ_TRANSPORTS.IPC,
        forked: bool = False,  # used by decorator
        **kwargs: typing.Dict[str, typing.Any],
    ) -> None:
        """
        Quick-start to serve `Thing` over ZMQ. This method is fully blocking.
        Call `exit()` to resume (untested).

        Parameters
        ----------
        access_points: list[ZMQ_TRANSPORTS] | ZMQ_TRANSPORTS | str | list[str], Default ZMQ_TRANSPORTS.IPC or "IPC"
            ZMQ transport layers at which the object is exposed:

            - `TCP` -  custom implemented protocol in plain TCP - supply a socket address additionally in the format
            `tcp://*:<port>` (for example - `tcp://*:5555`) or a random port will be automatically used.
            The star `*` indicates that the server will listen on all available network interfaces.
            - `IPC` - inter process communication - connection can be made from other processes running
            locally within same computer. No client on the network will be able to contact the object using
            this transport. Use this transport if you wish to avoid configuring your firewall for a local client.
            - `INPROC` - one main python process spawns several threads in one of which the `Thing`
            will be running. The object can be contacted by a client on another thread but not from other processes
            or the network.

            One may use more than one form of transport.  All requests made will be anyway queued internally
            irrespective of origin. For multiple transports, supply a list of transports. For example: `[ZMQ_TRANSPORTS.TCP, ZMQ_TRANSPORTS.IPC]`,
            `["TCP", "IPC"]`, `["tcp://*:5555", "IPC"]` or `["IPC", "INPROC"]`.

        forked: bool, Default `False`
            if `True`, the server is started in a separate thread and this method returns immediately.

        **kwargs:
            - context: `zmq.asyncio.Context`, optional,
                ZMQ context object to be used for creating sockets. If not supplied, a global shared context is used.
                For INPROC, either do not supply context or use the same context across all threads.
        """
        from .zmq.rpc_server import prepare_rpc_server

        prepare_rpc_server(instance=self, access_points=access_points, **kwargs)
        self.rpc_server.run()

    @forkable  # noqa: F405
    def run_with_http_server(
        self,
        port: int = 8080,
        address: str = "0.0.0.0",
        # host: str = None,
        allowed_clients: str | typing.Iterable[str] | None = None,
        ssl_context: ssl.SSLContext | None = None,
        # protocol_version : int = 1,
        # network_interface : str = 'Ethernet',
        forked: bool = False,  # used by forkable decorator
        **kwargs: typing.Dict[str, typing.Any],
    ) -> None:
        """
        Quick-start to serve `Thing` over HTTP. This method is fully blocking.

        Parameters
        ----------
        port: int
            the port at which the HTTP server should be run (unique)
        address: str
            A convenience option to set IP address apart from 0.0.0.0 (i.e. bind to all interfaces, which is default)
        ssl_context: ssl.SSLContext | None
            provide custom certificates with an SSL context for encrypted communication
        allowed_clients: typing.Iterable[str] | str | None
            serves request and sets CORS only from these clients, other clients are rejected with 403. Uses remote IP
            header value to achieve this. Unlike CORS, the server resource is not even executed if the client is not an allowed client.
            Note that the remote IP in a HTTP request is believable only from a trusted HTTP client, not a modified one.
        forked: bool, Default `False`
            if `True`, the server is started in a separate thread and this method returns immediately
        **kwargs: typing.Dict[str, typing.Any]
            additional keyword arguments:

            - `property_handler`: `BaseHandler` | `PropertyHandler`,
                custom web request handler for property operations
            - `action_handler`: `BaseHandler` | `ActionHandler`,
                custom web request handler for action operations
            - `event_handler`: `BaseHandler` | `EventHandler`,
                custom event handler for handling events
        """
        # network_interface: str
        #     Currently there is no logic to detect the IP addresss (as externally visible) correctly, therefore please
        #     send the network interface name to retrieve the IP. If a DNS server is present, you may leave this field
        # host: str
        #     Host Server to subscribe to coordinate starting sequence of things & web GUI
        from ..server.http import HTTPServer

        self.run_with_zmq_server(access_points=ZMQ_TRANSPORTS.INPROC, forked=True)
        http_server = HTTPServer(
            port=port,
            address=address,
            logger=self.logger,
            ssl_context=ssl_context,
            allowed_clients=allowed_clients,
            # network_interface=network_interface,
            **kwargs,
        )
        http_server.add_things(dict(INPROC=self.id))
        http_server.listen()

    @forkable  # noqa: F405
    def run(
        self,
        **kwargs: typing.Dict[str, typing.Any],
    ) -> None:
        """
        Expose the object with the given servers. This method is blocking until `exit()` is called.

        Parameters
        ----------
        kwargs: typing.Dict[str, Any]
            keyword arguments

            - `access_points`: dict[str, dict | int | str | list[str]], optional
                The protocol name and its port or parameters to expose the object.
                for example - `[('HTTP', 9000), ('ZMQ', 'tcp://*:5555')]`

            - `servers`: list[BaseProtocolServer]
                list of instantiated servers to expose the object.
        """
        from ..server.http import HTTPServer
        from ..server.zmq import ZMQServer
        from ..server.mqtt import MQTTPublisher
        from .zmq.rpc_server import RPCServer, prepare_rpc_server

        access_points = kwargs.get("access_points", None)  # type: dict[str, dict | int | str | list[str]]
        servers = kwargs.get("servers", [])  # type: typing.Optional[typing.List[BaseProtocolServer]]

        if access_points is None and len(servers) == 0:
            raise ValueError("At least one of access_points or servers must be provided.")
        if access_points is not None and len(servers) > 0:
            raise ValueError("Only one of access_points or servers can be provided.")

        if access_points is not None:
            for protocol, params in access_points:
                if protocol.upper() == "HTTP":
                    if isinstance(params, int):
                        params = dict(port=params)
                    http_server = HTTPServer(**params)
                    servers.append(http_server)
                elif protocol.upper() == "ZMQ":
                    if isinstance(params, int):
                        params = dict(access_points=[f"tcp://*:{params}"])
                    elif isinstance(params, str):
                        params = dict(access_points=[params])
                    elif isinstance(params, list):
                        params = dict(access_points=params)
                    if not any(
                        isinstance(ap, str) and ap.upper().startswith("INPROC")
                        for ap in params.get("access_points", [])
                    ):
                        params["access_points"].append("INPROC")
                    prepare_rpc_server(self, **params)
                    servers.append(self.rpc_server)
                else:
                    warnings.warn(f"Unsupported protocol: {protocol}", category=UserWarning)

        if not any(isinstance(server, (RPCServer, ZMQServer)) for server in servers):
            prepare_rpc_server(self, access_points=ZMQ_TRANSPORTS.INPROC)
        for server in servers:
            # this cannot be merged with the loop below because self.rpc_server needs to be set first
            if isinstance(server, (RPCServer, ZMQServer)):
                self.rpc_server = server

        for server in servers:
            if isinstance(server, HTTPServer):

                def start_http_server(server: HTTPServer) -> None:
                    server.router.add_thing_instance_through_broker(
                        server_id=self.rpc_server.id,
                        thing_id=self.id,
                        access_point="INPROC",
                    )
                    server.listen()

                threading.Thread(target=start_http_server, args=(server,)).start()
            elif isinstance(server, MQTTPublisher):
                server.add_thing(
                    server_id=self.rpc_server.id,
                    thing_id=self.id,
                    access_point="INPROC",
                )
                server.start()
        self.rpc_server.run()

    @action()
    def exit(self) -> None:
        """Stop serving the object. This method usually needs to be called remotely"""
        if self.rpc_server is None:
            self.logger.debug("exit() called on a object that is not exposed yet.")
            return
        if self._owners:
            raise NotImplementedError(
                "call exit on the top-level object, composed objects cannot exit the loop. "
                + f"This object belongs to {self._owners.__class__.__name__} with ID {self._owners.id}."
            )
        self.rpc_server.stop()

    @action()
    def ping(self) -> None:
        """
        ping to see if it is alive. Successful when action succeeds with no return value and
        no timeout or exception raised on the client side.
        """
        pass

    def __hash__(self) -> int:
        filename = inspect.getfile(self.__class__)
        if filename is not None:
            # i.e. try to make it as unique as possible
            return hash(filename + self.__class__.__name__ + self.id)
        return hash(self.__class__.__name__ + self.id)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Thing):
            return False
        return self.__class__ == other.__class__ and self.id == other.id

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.id})"

    def __contains__(self, item: Property | BoundAction | EventDispatcher) -> bool:
        return item in self.properties or item in self.actions or item in self.events

    def __enter__(self) -> "Thing":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        pass


from .state_machine import StateMachine  # noqa: F401, E402

__all__ = [Thing.__name__]



import typing
import warnings
import jsonschema
from enum import Enum
from types import FunctionType, MethodType
from inspect import iscoroutinefunction, getfullargspec
from pydantic import BaseModel, RootModel

from ..param.parameterized import ParameterizedFunction
from ..constants import JSON
from ..config import global_config
from ..utils import (
    get_return_type_from_signature,
    has_async_def,
    get_input_model_from_signature,
    issubklass,
    isclassmethod,
)
from ..exceptions import StateMachineError
from ..schema_validators.validators import JSONSchemaValidator, PydanticSchemaValidator
from .dataklasses import ActionInfoValidator


class Action:
    """
    Object that models an action.
    These actions are unbound and return a bound action when accessed using the owning object.
    """

    __slots__ = ["obj", "owner", "_execution_info"]

    def __init__(self, obj: MethodType) -> None:
        """
        Parameters
        ----------
        obj: MethodType
            the method that is being wrapped as an action
        """
        self.obj = obj

    def __set_name__(self, owner, name):
        self.owner = owner

    def __str__(self) -> str:
        return f"<Action({self.owner.__name__}.{self.obj.__name__})>"

    def __eq__(self, other) -> bool:
        if not isinstance(other, Action):
            return False
        return self.obj == other.obj

    def __hash__(self) -> int:
        return hash(self.obj)

    def __get__(self, instance, owner):
        if instance is None and not self._execution_info.isclassmethod:
            return self
        if self._execution_info.iscoroutine:
            return BoundAsyncAction(self.obj, self, instance, owner)
        return BoundSyncAction(self.obj, self, instance, owner)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError(
            f"Cannot invoke unbound action {self.name} of {self.owner.__name__}."
            + " Bound methods must be called, not the action itself. Use the appropriate instance to call the method."
        )

    @property
    def name(self) -> str:
        """name of the action"""
        return self.obj.__name__

    @property
    def execution_info(self) -> ActionInfoValidator:
        """
        internal dataclass that holds all information about the action

        TODO: this can be refactored
        """
        return self._execution_info

    @execution_info.setter
    def execution_info(self, value: ActionInfoValidator) -> None:
        if not isinstance(value, ActionInfoValidator):
            raise TypeError("execution_info must be of type ActionInfoValidator")
        self._execution_info = value  # type: ActionInfoValidator

    def to_affordance(self, owner_inst=None):
        """
        Generates a `ActionAffordance` TD fragment for this Action.

        Parameters
        ----------
        owner_inst: Thing, optional
            The instance of the owning `Thing` object. If not supplied, the class is used.

        Returns
        -------
        ActionAffordance
            the affordance TD fragment for this action
        """
        from ..td import ActionAffordance

        return ActionAffordance.generate(self, owner_inst or self.owner)


class BoundAction:
    """
    A bound action - base class for both sync and async methods.
    """

    __slots__ = [
        "obj",
        "execution_info",
        "descriptor",
        "owner_inst",
        "owner",
        "bound_obj",
    ]

    def __init__(self, obj: FunctionType, descriptor: Action, owner_inst, owner) -> None:
        self.obj = obj
        self.descriptor = descriptor
        self.execution_info = descriptor._execution_info
        self.owner = owner
        self.owner_inst = owner_inst
        self.bound_obj = owner if self.execution_info.isclassmethod else owner_inst

    def __post_init__(self):
        # never called, neither possible to call, only type hinting
        from .thing import ThingMeta, Thing

        # owner class and instance
        self.owner: ThingMeta
        self.owner_inst: Thing
        self.obj: FunctionType
        # the validator that was used to accept user inputs to this action.
        # stored only for reference, hardly used.
        self._execution_info: ActionInfoValidator

    def validate_call(self, args, kwargs: typing.Dict[str, typing.Any]) -> None:
        """
        Validate the call to the action, like payload, state machine state etc.
        Errors are raised as exceptions.
        """
        if self.execution_info.isparameterized and len(args) > 0:
            raise RuntimeError("parameterized functions cannot have positional arguments")
        if self.owner_inst is None:
            return
        if self.execution_info.state is None or (
            hasattr(self.owner_inst, "state_machine")
            and self.owner_inst.state_machine.current_state in self.execution_info.state
        ):
            if self.execution_info.schema_validator is not None:
                self.execution_info.schema_validator.validate_method_call(args, kwargs)
        else:
            raise StateMachineError(
                "Thing '{}' is in '{}' state, however action can be executed only in '{}' state".format(
                    f"{self.owner.__class__}.{self.owner_inst.id}",
                    self.owner_inst.state,
                    self.execution_info.state,
                )
            )

    @property
    def name(self) -> str:
        """name of the action"""
        return self.obj.__name__

    def __call__(self, *args, **kwargs):
        raise NotImplementedError("call must be implemented by subclass")

    def external_call(self, *args, **kwargs):
        """validated call to the action with state machine and payload checks"""
        raise NotImplementedError("external_call must be implemented by subclass")

    def __str__(self):
        return f"<BoundAction({self.owner.__name__}.{self.obj.__name__} of {self.owner_inst.id})>"

    def __eq__(self, value):
        if not isinstance(value, BoundAction):
            return False
        return self.obj == value.obj

    def __hash__(self):
        return hash(str(self))

    def __getattribute__(self, name):
        "Emulate method_getset() in Objects/classobject.c"
        # https://docs.python.org/3/howto/descriptor.html#functions-and-methods
        if name == "__doc__":
            return self.obj.__doc__
        return super().__getattribute__(name)

    def to_affordance(self):
        """
        Generates a `ActionAffordance` TD fragment for this Action.

        Parameters
        ----------
        owner_inst: Thing, optional
            The instance of the owning `Thing` object. If not supplied, the class is used.

        Returns
        -------
        ActionAffordance
            the affordance TD fragment for this action
        """
        return Action.to_affordance(self.descriptor, self.owner_inst or self.owner)


class BoundSyncAction(BoundAction):
    """
    non async(io) action call. The call is passed to the method as-it-is to allow local
    invocation without state machine checks.
    """

    def external_call(self, *args, **kwargs):
        """validated call to the action with state machine and payload checks"""
        self.validate_call(args, kwargs)
        return self.__call__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if self.execution_info.isclassmethod:
            return self.obj(*args, **kwargs)
        return self.obj(self.bound_obj, *args, **kwargs)


class BoundAsyncAction(BoundAction):
    """
    async(io) action call. The call is passed to the method as-it-is to allow local
    invocation without state machine checks.
    """

    async def external_call(self, *args, **kwargs):
        """validated call to the action with state machine and payload checks"""
        self.validate_call(args, kwargs)
        return await self.__call__(*args, **kwargs)

    async def __call__(self, *args, **kwargs):
        if self.execution_info.isclassmethod:
            return await self.obj(*args, **kwargs)
        return await self.obj(self.bound_obj, *args, **kwargs)


__action_kw_arguments__ = ["safe", "idempotent", "synchronous"]


def action(
    input_schema: JSON | BaseModel | RootModel | None = None,
    output_schema: JSON | BaseModel | RootModel | None = None,
    state: str | Enum | None = None,
    **kwargs,
) -> Action:
    """
    decorate on your methods to make them accessible remotely or create 'actions' out of them.

    Parameters
    ----------
    input_schema: JSON
        schema for arguments to validate
    output_schema: JSON
        schema for return value, currently only used to inform clients which are supposed to validate on their own
    state: str | Tuple[str], optional
        state machine state under which the action can be executed. When not provided, the action can be executed
        under any state.
    **kwargs:
        additional keyword arguments to specify action characteristics:

        - `synchronous`: bool,
            indicate in thing description if action is synchronous (not long running or async) - completes in a deterministic
            (& usually) short period of time, default `True`
        - `threaded`: bool,
            indicate that a method should be run in a separate thread, default `False`.
            Alternative to synchronous for non-async methods.
        - `create_task`: bool,
            indicate that a method function should be run in a new task, default `True`.
            Alternative to synchronous for async methods.
        - `safe`: bool,
            indicate in thing description if action is safe to execute, default `False`
        - `idempotent`: bool,
            indicate in thing description if action is idempotent (for example, allows HTTP clients to cache return value),
            default `False`

    Returns
    -------
    Action
        returns the callable object wrapped in an `Action` object. When accessed at instance level,
        a `BoundSyncAction` or `BoundAsyncAction` object is returned.
    """

    def inner(obj):
        input_schema = inner._arguments.get("input_schema", None)
        output_schema = inner._arguments.get("output_schema", None)
        state = inner._arguments.get("state", None)
        kwargs = inner._arguments.get("kwargs", {})

        original = obj
        if (
            not isinstance(obj, (FunctionType, MethodType, Action, BoundAction))
            and not isclassmethod(obj)
            and not issubklass(obj, ParameterizedFunction)
        ):
            raise TypeError(f"target for action or is not a function/method. Given type {type(obj)}") from None
        if isclassmethod(obj):
            obj = obj.__func__
        if isinstance(obj, (Action, BoundAction)):
            if obj.execution_info.isclassmethod:
                raise RuntimeError("cannot wrap a classmethod as action once again, please skip")
            warnings.warn(
                f"{obj.name} is already wrapped as an action, wrapping it again with newer settings.",
                category=UserWarning,
            )
            obj = obj.obj
        if obj.__name__.startswith("__"):
            raise ValueError(f"dunder objects cannot become remote : {obj.__name__}")
        execution_info_validator = ActionInfoValidator()
        if state is not None:
            if isinstance(state, (Enum, str)):
                execution_info_validator.state = (state,)
            else:
                execution_info_validator.state = state
        if "request" in getfullargspec(obj).kwonlyargs:
            execution_info_validator.request_as_argument = True
        execution_info_validator.isaction = True
        execution_info_validator.obj = original
        execution_info_validator.create_task = kwargs.get("create_task", False)
        execution_info_validator.safe = kwargs.get("safe", False)
        execution_info_validator.idempotent = kwargs.get("idempotent", False)
        execution_info_validator.synchronous = kwargs.get("synchronous", True)

        if isclassmethod(original):
            execution_info_validator.iscoroutine = has_async_def(obj)
            execution_info_validator.isclassmethod = True
        elif issubklass(obj, ParameterizedFunction):
            execution_info_validator.iscoroutine = iscoroutinefunction(obj.__call__)
            execution_info_validator.isparameterized = True
        else:
            execution_info_validator.iscoroutine = iscoroutinefunction(obj)

        if not input_schema:
            try:
                input_schema = get_input_model_from_signature(obj, remove_first_positional_arg=True)
            except Exception as ex:
                if global_config.VALIDATE_SCHEMAS:
                    warnings.warn(
                        f"Could not infer input schema for {obj.__name__} due to {str(ex)}. "
                        + "Considering filing a bug report if you think this should have worked correctly",
                        category=RuntimeWarning,
                    )
        if global_config.VALIDATE_SCHEMAS and input_schema:
            if isinstance(input_schema, dict):
                execution_info_validator.schema_validator = JSONSchemaValidator(input_schema)
            elif issubklass(input_schema, (BaseModel, RootModel)):
                execution_info_validator.schema_validator = PydanticSchemaValidator(input_schema)
            else:
                raise TypeError(
                    "input schema must be a JSON schema or a Pydantic model, got {}".format(type(input_schema))
                )
        execution_info_validator.argument_schema = input_schema

        if not output_schema:
            try:
                output_schema = get_return_type_from_signature(obj)
            except Exception as ex:
                warnings.warn(
                    f"Could not infer output schema for {obj.__name__} due to {str(ex)}. "
                    + "Considering filing a bug report if you think this should have worked correctly",
                    category=RuntimeWarning,
                )

        if output_schema:
            # output is not validated by us, so we just check the schema and dont create a validator
            if isinstance(output_schema, dict):
                jsonschema.Draft7Validator.check_schema(output_schema)
                execution_info_validator.return_value_schema = output_schema
            elif issubklass(output_schema, (BaseModel, RootModel)):
                execution_info_validator.return_value_schema = output_schema
            else:
                raise TypeError(
                    "output schema must be a JSON schema or a Pydantic model, got {}".format(type(output_schema))
                )

        final_obj = Action(original)  # type: Action
        final_obj.execution_info = execution_info_validator
        return final_obj

    if callable(input_schema):
        raise TypeError(
            "input schema should be a JSON or pydantic BaseModel, not a function/method, "
            + "did you decorate your action wrongly? use @action() instead of @action"
        )
    if any(key not in __action_kw_arguments__ for key in kwargs.keys()):
        raise ValueError(
            "Only 'safe', 'idempotent', 'synchronous' are allowed as keyword arguments, "
            + f"unknown arguments found {kwargs.keys()}"
        )
    inner._arguments = dict(
        input_schema=input_schema,
        output_schema=output_schema,
        state=state,
        kwargs=kwargs,
    )
    return inner


__all__ = [action.__name__, Action.__name__]



import asyncio
import typing
import zmq
import zmq.asyncio

from ..constants import ZMQ_TRANSPORTS
from ..utils import get_current_async_loop
from ..core.thing import Thing
from ..core.zmq.brokers import AsyncEventConsumer, AsyncZMQServer, EventPublisher
from ..core.zmq.rpc_server import RPCServer


class ZMQServer(RPCServer):
    """
    Server to expose `Thing` over `ZeroMQ` protocol. Extends `RPCServer` to support `IPC` & `TCP`.
    """

    def __init__(
        self,
        *,
        id: str,
        things: typing.List["Thing"],
        context: zmq.asyncio.Context | None = None,
        access_points: ZMQ_TRANSPORTS = ZMQ_TRANSPORTS.IPC,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        id : str
            Unique identifier for the server instance.
        things : List[Thing]
            List of `Thing` instances to be managed by the server.
        context : zmq.asyncio.Context, optional
            ZeroMQ context for socket management. If `None`, a global context is used.
        access_points : ZMQ_TRANSPORTS or List[ZMQ_TRANSPORTS], default ZMQ_TRANSPORTS.IPC
            Transport protocols for communication. Supported values are `ZMQ_TRANSPORTS.IPC`, `ZMQ_TRANSPORTS.TCP` or
            a TCP socket address `tcp://*:<port>`. Can be a single value or a list of values.
        **kwargs
            Additional keyword arguments for server configuration. Usually:

            - `logger`: `logging.Logger`, custom logger instance.
            - `poll_timeout`: `int`, polling timeout in milliseconds.
        """
        self.ipc_server = self.tcp_server = None
        self.ipc_event_publisher = self.tcp_event_publisher = self.inproc_events_proxy = None
        super().__init__(id=id, things=things, context=context, **kwargs)

        tcp_socket_address = None

        if isinstance(access_points, str):
            access_points = [access_points]
        elif not isinstance(access_points, list):
            raise TypeError(f"unsupported transport type : {type(access_points)}")
        for index, transport in enumerate(access_points):
            if isinstance(transport, str) and len(transport) in [3, 6]:
                access_points[index] = transport.upper()
            elif transport.lower().startswith("tcp://"):
                access_points[index] = "TCP"
                tcp_socket_address = transport
            else:
                access_points[index] = transport

        # initialise every externally visible protocol
        if ZMQ_TRANSPORTS.TCP in access_points or "TCP" in access_points:
            self.tcp_server = AsyncZMQServer(
                id=self.id,
                context=self.context,
                access_point=tcp_socket_address or ZMQ_TRANSPORTS.TCP,
                **kwargs,
            )
            host, port = self.tcp_server.socket_address.rsplit(":", 1)
            new_port = int(port) + 1  # try the next port for the event publisher
            tcp_socket_address = f"{host}:{new_port}"
            self.tcp_event_publisher = EventPublisher(
                id=f"{self.id}/event-publisher",
                context=self.context,
                transport=ZMQ_TRANSPORTS.TCP,
                access_point=tcp_socket_address,
                **kwargs,
            )
        if ZMQ_TRANSPORTS.IPC in access_points or "IPC" in access_points:
            self.ipc_server = AsyncZMQServer(
                id=self.id,
                context=self.context,
                access_point=ZMQ_TRANSPORTS.IPC,
                **kwargs,
            )
            self.ipc_event_publisher = EventPublisher(
                id=f"{self.id}/event-publisher",
                context=self.context,
                access_point=ZMQ_TRANSPORTS.IPC,
                **kwargs,
            )
        if self.ipc_event_publisher is not None or self.tcp_event_publisher is not None:
            self.inproc_events_proxy = AsyncEventConsumer(
                id=f"{self.id}/event-proxy",
                event_unique_identifier="",
                access_point=self.event_publisher.socket_address,
                context=self.context,
                **kwargs,
            )

    def run_zmq_request_listener(self) -> None:
        # doc in parent class
        eventloop = get_current_async_loop()
        if self.ipc_server is not None:
            eventloop.call_soon(lambda: asyncio.create_task(self.recv_requests_and_dispatch_jobs(self.ipc_server)))
        if self.tcp_server is not None:
            eventloop.call_soon(lambda: asyncio.create_task(self.recv_requests_and_dispatch_jobs(self.tcp_server)))
        if self.inproc_events_proxy is not None:
            eventloop.call_soon(lambda: asyncio.create_task(self.tunnel_events_from_inproc()))
        super().run_zmq_request_listener()

    async def tunnel_events_from_inproc(self) -> None:
        if not self.inproc_events_proxy:
            return
        self.inproc_events_proxy.subscribe()
        while self._run:
            try:
                event = await self.inproc_events_proxy.receive(raise_interrupt_as_exception=True)
                if not event:
                    continue
                if self.ipc_event_publisher is not None:
                    self.ipc_event_publisher.socket.send_multipart(event.byte_array)
                    # print(f"sent event to ipc publisher {event.byte_array}")
                if self.tcp_event_publisher is not None:
                    self.tcp_event_publisher.socket.send_multipart(event.byte_array)
                    # print(f"sent event to tcp publisher {event.byte_array}")
            except ConnectionAbortedError:
                break
            except Exception as e:
                self.logger.error(f"error in tunneling events from inproc: {e}")

    def stop(self) -> None:
        if self.ipc_server is not None:
            self.ipc_server.stop_polling()
        if self.tcp_server is not None:
            self.tcp_server.stop_polling()
        if self.inproc_events_proxy is not None:
            self.inproc_events_proxy.stop()
        super().stop()

    def exit(self) -> None:
        try:
            self.stop()
            if self.ipc_server is not None:
                self.ipc_server.exit()
                self.ipc_event_publisher.exit()
            if self.tcp_server is not None:
                self.tcp_server.exit()
                self.tcp_event_publisher.exit()
            if self.req_rep_server is not None:
                self.req_rep_server.exit()
            if self.event_publisher is not None:
                self.event_publisher.exit()
            if self.inproc_events_proxy is not None:
                self.inproc_events_proxy.exit()
        except Exception as ex:
            self.logger.warning(f"Exception occurred while exiting the server - {str(ex)}")

    def __str__(self):
        parts = [f"ZMQServer(\n\tid: {self.id}"]
        for name in [
            "ipc_server",
            "tcp_server",
            "req_rep_server",
            "ipc_event_publisher",
            "tcp_event_publisher",
            "event_publisher",
            "inproc_events_proxy",
        ]:
            obj = getattr(self, name, None)
            if obj is not None:
                type_name = type(obj).__name__
                parts.append(f"{name}: {getattr(obj, 'socket_address', None)} ({type_name})")
            else:
                parts.append(f"{name}: None")
        paths = "\n\t".join(parts)
        paths += "\n)"
        return paths



import unittest
import zmq.asyncio

from hololinked.core.zmq.brokers import BaseZMQ
from hololinked.constants import ZMQ_TRANSPORTS

try:
    from .utils import TestCase, TestRunner
except ImportError:
    from utils import TestCase, TestRunner


class TestSocket(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print(f"test ZMQ socket creation with {cls.__name__}")

    def test_1_socket_creation_defaults(self):
        """check the default settings of socket creation - an IPC socket which is a ROUTER and async"""
        socket, socket_address = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=zmq.asyncio.Context(),
        )
        self.assertIsInstance(socket, zmq.asyncio.Socket)
        self.assertTrue(socket.getsockopt_string(zmq.IDENTITY) == "test-server")
        self.assertTrue(socket.socket_type == zmq.ROUTER)
        self.assertTrue(socket_address.startswith("ipc://"))
        self.assertTrue(socket_address.endswith(".ipc"))
        socket.close()

    def test_2_context_options(self):
        """
        Check that context and socket type are as expected.
        Async context should be used for async socket and sync context for sync socket.
        """
        context = zmq.Context()
        socket, _ = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
        )
        self.assertTrue(isinstance(socket, zmq.Socket))
        self.assertTrue(not isinstance(socket, zmq.asyncio.Socket))
        socket.close()
        context.term()

        context = zmq.asyncio.Context()
        socket, _ = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
        )
        self.assertTrue(isinstance(socket, zmq.Socket))
        self.assertTrue(isinstance(socket, zmq.asyncio.Socket))
        socket.close()
        context.term()

    def test_3_transport_options(self):
        """check only three transport options are supported"""
        context = zmq.asyncio.Context()
        socket, socket_address = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            access_point="tcp://*:5555",
        )
        for sock_addr in [socket_address, socket.getsockopt_string(zmq.LAST_ENDPOINT)]:
            self.assertTrue(sock_addr.startswith("tcp://"))
            self.assertTrue(sock_addr.endswith(":5555"))
        socket.close()

        socket, socket_address = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            access_point="IPC",
        )

        self.assertEqual(socket_address, socket.getsockopt_string(zmq.LAST_ENDPOINT))
        self.assertTrue(socket_address.startswith("ipc://"))
        self.assertTrue(socket_address.endswith(".ipc"))
        socket.close()

        socket, socket_address = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            access_point="INPROC",
        )
        self.assertEqual(socket_address, socket.getsockopt_string(zmq.LAST_ENDPOINT))
        self.assertTrue(socket_address.startswith("inproc://"))
        self.assertTrue(socket_address.endswith("test-server"))
        socket.close()
        context.term()

        # Specify transport as enum and do the same tests
        context = zmq.Context()
        socket, socket_address = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            access_point=ZMQ_TRANSPORTS.INPROC,
        )
        self.assertTrue(socket_address.startswith("inproc://"))
        self.assertTrue(socket_address.endswith("test-server"))
        socket.close()

        socket, socket_address = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            access_point=ZMQ_TRANSPORTS.IPC,
        )
        self.assertTrue(socket_address.startswith("ipc://"))
        self.assertTrue(socket_address.endswith(".ipc"))
        socket.close()

        socket, socket_address = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            access_point=ZMQ_TRANSPORTS.TCP,
        )
        self.assertTrue(socket_address.startswith("tcp://"))
        # Strip the port number from TCP address and check if it's a valid port integer
        host, port_str = socket_address.rsplit(":", 1)
        self.assertTrue(port_str.isdigit())
        self.assertTrue(0 < int(port_str) < 65536)
        socket.close()
        context.term()

        # check that other transport options raise error
        context = zmq.asyncio.Context()
        self.assertRaises(
            NotImplementedError,
            lambda: BaseZMQ.get_socket(
                server_id="test-server",
                socket_id="test-server",
                node_type="server",
                context=context,
                access_point="PUB",
            ),
        )
        context.term()

    def test_4_socket_options(self):
        """check that socket options are as expected"""
        context = zmq.asyncio.Context()

        socket, _ = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            socket_type=zmq.ROUTER,
        )
        self.assertTrue(socket.socket_type == zmq.ROUTER)
        self.assertTrue(socket.getsockopt_string(zmq.IDENTITY) == "test-server")
        self.assertTrue(isinstance(socket, zmq.asyncio.Socket))
        socket.close()

        socket, _ = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            socket_type=zmq.DEALER,
        )
        self.assertTrue(socket.socket_type == zmq.DEALER)
        self.assertTrue(socket.getsockopt_string(zmq.IDENTITY) == "test-server")
        self.assertTrue(isinstance(socket, zmq.asyncio.Socket))
        socket.close()

        socket, _ = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            socket_type=zmq.PUB,
        )
        self.assertTrue(socket.socket_type == zmq.PUB)
        self.assertTrue(socket.getsockopt_string(zmq.IDENTITY) == "test-server")
        self.assertTrue(isinstance(socket, zmq.asyncio.Socket))
        socket.close()

        socket, _ = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            socket_type=zmq.SUB,
        )
        self.assertTrue(socket.socket_type == zmq.SUB)
        self.assertTrue(socket.getsockopt_string(zmq.IDENTITY) == "test-server")
        self.assertTrue(isinstance(socket, zmq.asyncio.Socket))
        socket.close()

        socket, _ = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            socket_type=zmq.PAIR,
        )
        self.assertTrue(socket.socket_type == zmq.PAIR)
        self.assertTrue(socket.getsockopt_string(zmq.IDENTITY) == "test-server")
        self.assertTrue(isinstance(socket, zmq.asyncio.Socket))
        socket.close()

        socket, _ = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            socket_type=zmq.PUSH,
        )
        self.assertTrue(socket.socket_type == zmq.PUSH)
        self.assertTrue(socket.getsockopt_string(zmq.IDENTITY) == "test-server")
        self.assertTrue(isinstance(socket, zmq.asyncio.Socket))
        socket.close()

        socket, _ = BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            socket_type=zmq.PULL,
        )
        self.assertTrue(socket.socket_type == zmq.PULL)
        self.assertTrue(socket.getsockopt_string(zmq.IDENTITY) == "test-server")
        self.assertTrue(isinstance(socket, zmq.asyncio.Socket))
        socket.close()
        context.term()


if __name__ == "__main__":
    unittest.main(testRunner=TestRunner())


"""
TODO:
1. check node_type values
2. check if TCP socket search happens
"""



import asyncio
import unittest
import logging

from hololinked.utils import isclassmethod
from hololinked.core.actions import (
    Action,
    BoundAction,
    BoundSyncAction,
    BoundAsyncAction,
)
from hololinked.core.dataklasses import ActionInfoValidator
from hololinked.core.thing import Thing, action
from hololinked.td.interaction_affordance import ActionAffordance
from hololinked.schema_validators import JSONSchemaValidator

try:
    from .utils import TestCase, TestRunner
    from .things import TestThing
    from .things.test_thing import replace_methods_with_actions
except ImportError:
    from utils import TestCase, TestRunner
    from things import TestThing
    from things.test_thing import replace_methods_with_actions


class TestAction(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print(f"test action with {cls.__name__}")

    def test_1_allowed_actions(self):
        """Test if methods can be decorated with action"""
        # 1. instance method can be decorated with action
        self.assertEqual(TestThing.action_echo, action()(TestThing.action_echo.obj))  # already predecorated as action
        # 2. classmethod can be decorated with action
        self.assertEqual(
            Action(TestThing.action_echo_with_classmethod),
            action()(TestThing.action_echo_with_classmethod),
        )
        self.assertTrue(isclassmethod(TestThing.action_echo_with_classmethod))
        # 3. async methods can be decorated with action
        self.assertEqual(Action(TestThing.action_echo_async), action()(TestThing.action_echo_async))
        # 4. async classmethods can be decorated with action
        self.assertEqual(
            Action(TestThing.action_echo_async_with_classmethod),
            action()(TestThing.action_echo_async_with_classmethod),
        )
        self.assertTrue(isclassmethod(TestThing.action_echo_async_with_classmethod))
        # 5. parameterized function can be decorated with action
        self.assertEqual(
            Action(TestThing.parameterized_action),
            action(safe=True)(TestThing.parameterized_action),
        )
        self.assertEqual(
            Action(TestThing.parameterized_action_without_call),
            action(idempotent=True)(TestThing.parameterized_action_without_call),
        )
        self.assertEqual(
            Action(TestThing.parameterized_action_async),
            action(synchronous=True)(TestThing.parameterized_action_async),
        )
        # 6. actions with input and output schema
        self.assertEqual(
            Action(TestThing.json_schema_validated_action),
            action(
                input_schema={
                    "val1": "integer",
                    "val2": "string",
                    "val3": "object",
                    "val4": "array",
                },
                output_schema={"val1": "int", "val3": "dict"},
            )(TestThing.json_schema_validated_action),
        )
        self.assertEqual(
            Action(TestThing.pydantic_validated_action),
            action()(TestThing.pydantic_validated_action),
        )

    def test_2_bound_method(self):
        """Test if methods decorated with action are correctly bound"""
        thing = TestThing(id="test-action", log_level=logging.ERROR)
        replace_methods_with_actions(thing_cls=TestThing)

        # 1. instance method can be decorated with action
        self.assertIsInstance(thing.action_echo, BoundAction)
        self.assertIsInstance(thing.action_echo, BoundSyncAction)
        self.assertNotIsInstance(thing.action_echo, BoundAsyncAction)
        self.assertIsInstance(TestThing.action_echo, Action)
        self.assertNotIsInstance(TestThing.action_echo, BoundAction)
        # associated attributes of BoundAction
        assert isinstance(thing.action_echo, BoundAction)  # type definition
        self.assertEqual(thing.action_echo.name, "action_echo")
        self.assertEqual(thing.action_echo.owner_inst, thing)
        self.assertEqual(thing.action_echo.owner, TestThing)
        self.assertEqual(thing.action_echo.execution_info, TestThing.action_echo.execution_info)
        self.assertEqual(
            str(thing.action_echo),
            f"<BoundAction({TestThing.__name__}.{thing.action_echo.name} of {thing.id})>",
        )
        self.assertNotEqual(thing.action_echo, TestThing.action_echo)
        self.assertEqual(thing.action_echo.bound_obj, thing)

        # 2. classmethod can be decorated with action
        self.assertIsInstance(thing.action_echo_with_classmethod, BoundAction)
        self.assertIsInstance(thing.action_echo_with_classmethod, BoundSyncAction)
        self.assertNotIsInstance(thing.action_echo_with_classmethod, BoundAsyncAction)
        self.assertIsInstance(TestThing.action_echo_with_classmethod, BoundAction)
        self.assertIsInstance(TestThing.action_echo_with_classmethod, BoundSyncAction)
        self.assertNotIsInstance(TestThing.action_echo_with_classmethod, Action)
        # associated attributes of BoundAction
        assert isinstance(thing.action_echo_with_classmethod, BoundAction)
        self.assertEqual(thing.action_echo_with_classmethod.name, "action_echo_with_classmethod")
        self.assertEqual(thing.action_echo_with_classmethod.owner_inst, thing)
        self.assertEqual(thing.action_echo_with_classmethod.owner, TestThing)
        self.assertEqual(
            thing.action_echo_with_classmethod.execution_info,
            TestThing.action_echo_with_classmethod.execution_info,
        )
        self.assertEqual(
            str(thing.action_echo_with_classmethod),
            f"<BoundAction({TestThing.__name__}.{thing.action_echo_with_classmethod.name} of {thing.id})>",
        )
        self.assertEqual(thing.action_echo_with_classmethod, TestThing.action_echo_with_classmethod)
        self.assertEqual(thing.action_echo_with_classmethod.bound_obj, TestThing)

        # 3. async methods can be decorated with action
        self.assertIsInstance(thing.action_echo_async, BoundAction)
        self.assertNotIsInstance(thing.action_echo_async, BoundSyncAction)
        self.assertIsInstance(thing.action_echo_async, BoundAsyncAction)
        self.assertIsInstance(TestThing.action_echo_async, Action)
        self.assertNotIsInstance(TestThing.action_echo_async, BoundAction)
        # associated attributes of BoundAction
        assert isinstance(thing.action_echo_async, BoundAction)
        self.assertEqual(thing.action_echo_async.name, "action_echo_async")
        self.assertEqual(thing.action_echo_async.owner_inst, thing)
        self.assertEqual(thing.action_echo_async.owner, TestThing)
        self.assertEqual(
            thing.action_echo_async.execution_info,
            TestThing.action_echo_async.execution_info,
        )
        self.assertEqual(
            str(thing.action_echo_async),
            f"<BoundAction({TestThing.__name__}.{thing.action_echo_async.name} of {thing.id})>",
        )
        self.assertNotEqual(thing.action_echo_async, TestThing.action_echo_async)
        self.assertEqual(thing.action_echo_async.bound_obj, thing)

        # 4. async classmethods can be decorated with action
        self.assertIsInstance(thing.action_echo_async_with_classmethod, BoundAction)
        self.assertNotIsInstance(thing.action_echo_async_with_classmethod, BoundSyncAction)
        self.assertIsInstance(thing.action_echo_async_with_classmethod, BoundAsyncAction)
        self.assertIsInstance(TestThing.action_echo_async_with_classmethod, BoundAction)
        self.assertIsInstance(TestThing.action_echo_async_with_classmethod, BoundAsyncAction)
        self.assertNotIsInstance(TestThing.action_echo_async_with_classmethod, Action)
        # associated attributes of BoundAction
        assert isinstance(thing.action_echo_async_with_classmethod, BoundAction)
        self.assertEqual(
            thing.action_echo_async_with_classmethod.name,
            "action_echo_async_with_classmethod",
        )
        self.assertEqual(thing.action_echo_async_with_classmethod.owner_inst, thing)
        self.assertEqual(thing.action_echo_async_with_classmethod.owner, TestThing)
        self.assertEqual(
            thing.action_echo_async_with_classmethod.execution_info,
            TestThing.action_echo_async_with_classmethod.execution_info,
        )
        self.assertEqual(
            str(thing.action_echo_async_with_classmethod),
            f"<BoundAction({TestThing.__name__}.{thing.action_echo_async_with_classmethod.name} of {thing.id})>",
        )
        self.assertEqual(
            thing.action_echo_async_with_classmethod,
            TestThing.action_echo_async_with_classmethod,
        )
        self.assertEqual(thing.action_echo_async_with_classmethod.bound_obj, TestThing)

        # 5. parameterized function can be decorated with action
        self.assertIsInstance(thing.parameterized_action, BoundAction)
        self.assertIsInstance(thing.parameterized_action, BoundSyncAction)
        self.assertNotIsInstance(thing.parameterized_action, BoundAsyncAction)
        self.assertIsInstance(TestThing.parameterized_action, Action)
        self.assertNotIsInstance(TestThing.parameterized_action, BoundAction)
        # associated attributes of BoundAction
        assert isinstance(thing.parameterized_action, BoundAction)
        self.assertEqual(thing.parameterized_action.name, "parameterized_action")
        self.assertEqual(thing.parameterized_action.owner_inst, thing)
        self.assertEqual(thing.parameterized_action.owner, TestThing)
        self.assertEqual(
            thing.parameterized_action.execution_info,
            TestThing.parameterized_action.execution_info,
        )
        self.assertEqual(
            str(thing.parameterized_action),
            f"<BoundAction({TestThing.__name__}.{thing.parameterized_action.name} of {thing.id})>",
        )
        self.assertNotEqual(thing.parameterized_action, TestThing.parameterized_action)
        self.assertEqual(thing.parameterized_action.bound_obj, thing)

        # 6. parameterized function can be decorated with action
        self.assertIsInstance(thing.parameterized_action_without_call, BoundAction)
        self.assertIsInstance(thing.parameterized_action_without_call, BoundSyncAction)
        self.assertNotIsInstance(thing.parameterized_action_without_call, BoundAsyncAction)
        self.assertIsInstance(TestThing.parameterized_action_without_call, Action)
        self.assertNotIsInstance(TestThing.parameterized_action_without_call, BoundAction)
        # associated attributes of BoundAction
        assert isinstance(thing.parameterized_action_without_call, BoundAction)
        self.assertEqual(
            thing.parameterized_action_without_call.name,
            "parameterized_action_without_call",
        )
        self.assertEqual(thing.parameterized_action_without_call.owner_inst, thing)
        self.assertEqual(thing.parameterized_action_without_call.owner, TestThing)
        self.assertEqual(
            thing.parameterized_action_without_call.execution_info,
            TestThing.parameterized_action_without_call.execution_info,
        )
        self.assertEqual(
            str(thing.parameterized_action_without_call),
            f"<BoundAction({TestThing.__name__}.{thing.parameterized_action_without_call.name} of {thing.id})>",
        )
        self.assertNotEqual(
            thing.parameterized_action_without_call,
            TestThing.parameterized_action_without_call,
        )
        self.assertEqual(thing.parameterized_action_without_call.bound_obj, thing)

        # 7. parameterized function can be decorated with action
        self.assertIsInstance(thing.parameterized_action_async, BoundAction)
        self.assertNotIsInstance(thing.parameterized_action_async, BoundSyncAction)
        self.assertIsInstance(thing.parameterized_action_async, BoundAsyncAction)
        self.assertIsInstance(TestThing.parameterized_action_async, Action)
        self.assertNotIsInstance(TestThing.parameterized_action_async, BoundAction)
        # associated attributes of BoundAction
        assert isinstance(thing.parameterized_action_async, BoundAction)
        self.assertEqual(thing.parameterized_action_async.name, "parameterized_action_async")
        self.assertEqual(thing.parameterized_action_async.owner_inst, thing)
        self.assertEqual(thing.parameterized_action_async.owner, TestThing)
        self.assertEqual(
            thing.parameterized_action_async.execution_info,
            TestThing.parameterized_action_async.execution_info,
        )
        self.assertEqual(
            str(thing.parameterized_action_async),
            f"<BoundAction({TestThing.__name__}.{thing.parameterized_action_async.name} of {thing.id})>",
        )
        self.assertNotEqual(thing.parameterized_action_async, TestThing.parameterized_action_async)
        self.assertEqual(thing.parameterized_action_async.bound_obj, thing)

        # 8. actions with input and output schema
        self.assertIsInstance(thing.json_schema_validated_action, BoundAction)
        self.assertIsInstance(thing.json_schema_validated_action, BoundSyncAction)
        self.assertNotIsInstance(thing.json_schema_validated_action, BoundAsyncAction)
        self.assertIsInstance(TestThing.json_schema_validated_action, Action)
        self.assertNotIsInstance(TestThing.json_schema_validated_action, BoundAction)
        # associated attributes of BoundAction
        assert isinstance(thing.json_schema_validated_action, BoundAction)
        self.assertEqual(thing.json_schema_validated_action.name, "json_schema_validated_action")
        self.assertEqual(thing.json_schema_validated_action.owner_inst, thing)
        self.assertEqual(thing.json_schema_validated_action.owner, TestThing)
        self.assertEqual(
            thing.json_schema_validated_action.execution_info,
            TestThing.json_schema_validated_action.execution_info,
        )
        self.assertEqual(
            str(thing.json_schema_validated_action),
            f"<BoundAction({TestThing.__name__}.{thing.json_schema_validated_action.name} of {thing.id})>",
        )
        self.assertNotEqual(thing.json_schema_validated_action, TestThing.json_schema_validated_action)
        self.assertEqual(thing.json_schema_validated_action.bound_obj, thing)

    def test_3_remote_info(self):
        """Test if the validator is working correctly, on which the logic of the action is based"""
        # basic check if the remote_info is correct, although this test is not necessary, not recommended and
        # neither particularly useful
        remote_info = TestThing.action_echo.execution_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)  # type definition
        self.assertTrue(remote_info.isaction)
        self.assertFalse(remote_info.isproperty)
        self.assertFalse(remote_info.isparameterized)
        self.assertFalse(remote_info.iscoroutine)
        self.assertFalse(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertTrue(remote_info.synchronous)

        remote_info = TestThing.action_echo_async.execution_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)  # type definition
        self.assertTrue(remote_info.isaction)
        self.assertTrue(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertFalse(remote_info.isparameterized)
        self.assertFalse(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertTrue(remote_info.synchronous)

        remote_info = TestThing.action_echo_with_classmethod.execution_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)  # type definition
        self.assertTrue(remote_info.isaction)
        self.assertFalse(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertFalse(remote_info.isparameterized)
        self.assertFalse(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertTrue(remote_info.synchronous)

        remote_info = TestThing.parameterized_action.execution_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)
        self.assertTrue(remote_info.isaction)
        self.assertFalse(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertTrue(remote_info.isparameterized)
        self.assertTrue(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertTrue(remote_info.synchronous)

        remote_info = TestThing.parameterized_action_without_call.execution_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)
        self.assertTrue(remote_info.isaction)
        self.assertFalse(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertTrue(remote_info.isparameterized)
        self.assertFalse(remote_info.safe)
        self.assertTrue(remote_info.idempotent)
        self.assertTrue(remote_info.synchronous)

        remote_info = TestThing.parameterized_action_async.execution_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)
        self.assertTrue(remote_info.isaction)
        self.assertTrue(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertTrue(remote_info.isparameterized)
        self.assertFalse(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertTrue(remote_info.synchronous)

        remote_info = TestThing.json_schema_validated_action.execution_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)
        self.assertTrue(remote_info.isaction)
        self.assertFalse(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertFalse(remote_info.isparameterized)
        self.assertFalse(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertTrue(remote_info.synchronous)
        self.assertIsInstance(remote_info.schema_validator, JSONSchemaValidator)

    def test_4_api_and_invalid_actions(self):
        """Test if action prevents invalid objects from being named as actions and raises neat errors"""
        # done allow action decorator to be terminated without '()' on a method
        with self.assertRaises(TypeError) as ex:
            action(TestThing.incorrectly_decorated_method)
        self.assertTrue(
            str(ex.exception).startswith(
                "input schema should be a JSON or pydantic BaseModel, not a function/method, did you decorate your action wrongly?"
            )
        )

        # dunder methods cannot be decorated with action
        with self.assertRaises(ValueError) as ex:
            action()(TestThing.__internal__)
        self.assertTrue(str(ex.exception).startswith("dunder objects cannot become remote"))

        # only functions and methods can be decorated with action
        for obj in [
            TestThing,
            str,
            1,
            1.0,
            "Str",
            True,
            None,
            object(),
            type,
            property,
        ]:
            with self.assertRaises(TypeError) as ex:
                action()(obj)  # not an action
            self.assertTrue(str(ex.exception).startswith("target for action or is not a function/method."))

        with self.assertRaises(ValueError) as ex:
            action(safe=True, some_kw=1)
        self.assertTrue(str(ex.exception).startswith("Only 'safe', 'idempotent', 'synchronous' are allowed"))

    # TODO - rename this test
    def test_5_thing_cls_actions(self):
        """Test class and instance level action access"""
        thing = TestThing(id="test-action", log_level=logging.ERROR)
        # class level
        for name, action in TestThing.actions.descriptors.items():
            self.assertIsInstance(action, Action)
        for name in replace_methods_with_actions._exposed_actions:
            self.assertTrue(name in TestThing.actions)
        # instance level
        for name, action in thing.actions.values.items():
            self.assertIsInstance(action, BoundAction)
        for name in replace_methods_with_actions._exposed_actions:
            self.assertTrue(name in thing.actions)
        # cannot call an instance bound action at class level
        self.assertRaises(NotImplementedError, lambda: TestThing.action_echo(thing, 1))
        # but can call instance bound action with instance
        self.assertEqual(1, thing.action_echo(1))
        # can also call classmethods as usual
        self.assertEqual(2, TestThing.action_echo_with_classmethod(2))
        self.assertEqual(3, thing.action_echo_with_classmethod(3))
        # async methods behave similarly
        self.assertEqual(4, asyncio.run(thing.action_echo_async(4)))
        self.assertEqual(5, asyncio.run(TestThing.action_echo_async_with_classmethod(5)))
        self.assertRaises(NotImplementedError, lambda: asyncio.run(TestThing.action_echo(7)))
        # parameterized actions behave similarly
        self.assertEqual(
            ("test-action", 1, "hello1", 1.1),
            thing.parameterized_action(1, "hello1", 1.1),
        )
        self.assertEqual(
            ("test-action", 2, "hello2", "foo2"),
            asyncio.run(thing.parameterized_action_async(2, "hello2", "foo2")),
        )
        self.assertRaises(NotImplementedError, lambda: TestThing.parameterized_action(3, "hello3", 5))
        self.assertRaises(
            NotImplementedError,
            lambda: asyncio.run(TestThing.parameterized_action_async(4, "hello4", 5)),
        )

    def test_6_action_affordance(self):
        """Test if action affordance is correctly created"""
        thing = TestThing(id="test-action", log_level=logging.ERROR)

        assert isinstance(thing.action_echo, BoundAction)  # type definition
        affordance = thing.action_echo.to_affordance()
        self.assertIsInstance(affordance, ActionAffordance)
        self.assertIsNone(affordance.idempotent)  # by default, not idempotent
        self.assertTrue(affordance.synchronous)  # by default, not synchronous
        self.assertIsNone(affordance.safe)  # by default, not safe
        self.assertIsNone(affordance.input)  # no input schema
        self.assertIsNone(affordance.output)  # no output schema
        self.assertIsNone(affordance.description)  # no doc

        assert isinstance(thing.action_echo_with_classmethod, BoundAction)  # type definition
        affordance = thing.action_echo_with_classmethod.to_affordance()
        self.assertIsInstance(affordance, ActionAffordance)
        self.assertIsNone(affordance.idempotent)  # by default, not idempotent
        self.assertTrue(affordance.synchronous)  # by default, synchronous
        self.assertIsNone(affordance.safe)  # by default, not safe
        self.assertIsNone(affordance.input)  # no input schema
        self.assertIsNone(affordance.output)  # no output schema
        self.assertIsNone(affordance.description)  # no doc

        assert isinstance(thing.action_echo_async, BoundAction)  # type definition
        affordance = thing.action_echo_async.to_affordance()
        self.assertIsInstance(affordance, ActionAffordance)
        self.assertIsNone(affordance.idempotent)  # by default, not idempotent
        self.assertTrue(affordance.synchronous)  # by default, synchronous
        self.assertIsNone(affordance.safe)  # by default, not safe
        self.assertIsNone(affordance.input)  # no input schema
        self.assertIsNone(affordance.output)  # no output schema
        self.assertIsNone(affordance.description)  # no doc

        assert isinstance(thing.action_echo_async_with_classmethod, BoundAction)  # type definition
        affordance = thing.action_echo_async_with_classmethod.to_affordance()
        self.assertIsInstance(affordance, ActionAffordance)
        self.assertIsNone(affordance.idempotent)  # by default, not idempotent
        self.assertTrue(affordance.synchronous)  # by default, synchronous
        self.assertIsNone(affordance.safe)  # by default, not safe
        self.assertIsNone(affordance.input)  # no input schema
        self.assertIsNone(affordance.output)  # no output schema
        self.assertIsNone(affordance.description)  # no doc

        assert isinstance(thing.parameterized_action, BoundAction)  # type definition
        affordance = thing.parameterized_action.to_affordance()
        self.assertIsInstance(affordance, ActionAffordance)
        self.assertIsNone(affordance.idempotent)
        self.assertTrue(affordance.synchronous)
        self.assertTrue(affordance.safe)
        self.assertIsNone(affordance.input)
        self.assertIsNone(affordance.output)
        self.assertIsNone(affordance.description)

        assert isinstance(thing.parameterized_action_without_call, BoundAction)  # type definition
        affordance = thing.parameterized_action_without_call.to_affordance()
        self.assertIsInstance(affordance, ActionAffordance)
        self.assertTrue(affordance.idempotent)  # by default, not idempotent
        self.assertTrue(affordance.synchronous)  # by default, synchronous
        self.assertIsNone(affordance.safe)  # by default, not safe
        self.assertIsNone(affordance.input)  # no input schema
        self.assertIsNone(affordance.output)  # no output schema
        self.assertIsNone(affordance.description)  # no doc

        assert isinstance(thing.parameterized_action_async, BoundAction)  # type definition
        affordance = thing.parameterized_action_async.to_affordance()
        self.assertIsInstance(affordance, ActionAffordance)
        self.assertIsNone(affordance.idempotent)  # by default, not idempotent
        self.assertTrue(affordance.synchronous)  # by default, not synchronous
        self.assertIsNone(affordance.safe)  # by default, not safe
        self.assertIsNone(affordance.input)  # no input schema
        self.assertIsNone(affordance.output)  # no output schema
        self.assertIsNone(affordance.description)  # no doc

        assert isinstance(thing.json_schema_validated_action, BoundAction)  # type definition
        affordance = thing.json_schema_validated_action.to_affordance()
        self.assertIsInstance(affordance, ActionAffordance)
        self.assertIsNone(affordance.idempotent)  # by default, not idempotent
        self.assertTrue(affordance.synchronous)  # by default, not synchronous
        self.assertIsNone(affordance.safe)  # by default, not safe
        self.assertIsInstance(affordance.input, dict)
        self.assertIsInstance(affordance.output, dict)
        self.assertIsNone(affordance.description)  # no doc


if __name__ == "__main__":
    unittest.main(testRunner=TestRunner())
